import pymimir as mm

from pathlib import Path

ROOT_DIR = (Path(__file__).parent.parent.parent.parent).absolute()


# def test_condition_grounder():
#     """ Test implementation of `LiftedConjunctionGrounder`.
#     """
#     domain_filepath = str(ROOT_DIR / "data" / "gripper" / "domain.pddl")
#     problem_filepath = str(ROOT_DIR / "data" / "gripper" / "test_problem.pddl")
#     parser = mm.PDDLParser(domain_filepath, problem_filepath)

#     domain = parser.get_domain()
#     problem = parser.get_problem()
#     pddl_repositories = parser.get_pddl_repositories()
#     goal_variables, goal_lifted, goal_static_lifted = problem.get_fluent_goal_condition().lift(domain, pddl_repositories)
#     grounder = mm.Grounder(parser.get_problem(), parser.get_pddl_repositories())
#     existentially_quantified_conjunctive_condition = pddl_repositories.get_or_create_existentially_quantified_conjunctive_condition(goal_variables, mm.StaticLiteralList(), goal_lifted, mm.DerivedLiteralList())
#     condition_grounder = mm.SatisficingBindingGenerator(grounder.get_literal_grounder(), existentially_quantified_conjunctive_condition)
#     applicable_action_generator = mm.LiftedApplicableActionGenerator(grounder.get_action_grounder())
#     axiom_evaluator = mm.LiftedAxiomEvaluator(grounder.get_axiom_grounder())
#     state_repository = mm.StateRepository(axiom_evaluator)
#     initial_state = state_repository.get_or_create_initial_state()
#     goal_bindings = condition_grounder.generate_ground_conjunctions(initial_state, 1_000_000)

#     del state_repository
#     del applicable_action_generator
#     del condition_grounder

#     assert goal_bindings
#     assert len(goal_bindings) == 2


if __name__ == "__main__":
    # test_condition_grounder()
    pass



""" This example illustrates the interface of the description logics component.
"""

import pymimir.advanced.search as search
import pymimir.advanced.datasets as datasets
import pymimir.advanced.languages.description_logics as description_logics

from functools import singledispatchmethod
from pathlib import Path

ROOT_DIR = (Path(__file__).parent.parent.parent).absolute()


class MyConstructorVisitor():
    """ Recursively visit the constructors. """
    def __init__(self):
        pass

    @singledispatchmethod
    def visit(self, arg):
        raise NotImplementedError("Missing implementation for the constructor!")

    @visit.register
    def _(self, constructor : description_logics.ConceptBot):
        pass
    @visit.register
    def _(self, constructor : description_logics.ConceptTop):
        pass
    @visit.register
    def _(self, constructor : description_logics.ConceptStaticAtomicState):
        pass
    @visit.register
    def _(self, constructor : description_logics.ConceptFluentAtomicState): 
        pass
    @visit.register
    def _(self, constructor : description_logics.ConceptDerivedAtomicState): 
        pass
    @visit.register
    def _(self, constructor : description_logics.ConceptStaticAtomicGoal): 
        pass
    @visit.register
    def _(self, constructor : description_logics.ConceptFluentAtomicGoal): 
        pass
    @visit.register
    def _(self, constructor : description_logics.ConceptDerivedAtomicGoal): 
        pass
    @visit.register
    def _(self, constructor : description_logics.ConceptIntersection): 
        self.visit(constructor.get_left_concept())
        self.visit(constructor.get_right_concept())
    @visit.register
    def _(self, constructor : description_logics.ConceptUnion): 
        self.visit(constructor.get_left_concept())
        self.visit(constructor.get_right_concept())
    @visit.register
    def _(self, constructor : description_logics.ConceptNegation): 
        self.visit(constructor.get_concept())
    @visit.register
    def _(self, constructor : description_logics.ConceptValueRestriction): 
        self.visit(constructor.get_role())
        self.visit(constructor.get_concept())
    @visit.register
    def _(self, constructor : description_logics.ConceptExistentialQuantification): 
        self.visit(constructor.get_role())
        self.visit(constructor.get_concept())
    @visit.register
    def _(self, constructor : description_logics.ConceptRoleValueMapContainment): 
        self.visit(constructor.get_left_role())
        self.visit(constructor.get_right_role())
    @visit.register
    def _(self, constructor : description_logics.ConceptRoleValueMapEquality): 
        self.visit(constructor.get_left_role())
        self.visit(constructor.get_right_role())
    @visit.register
    def _(self, constructor : description_logics.ConceptNominal): 
        pass
    @visit.register
    def _(self, constructor : description_logics.RoleUniversal): 
        pass
    @visit.register
    def _(self, constructor : description_logics.RoleStaticAtomicState): 
        pass
    @visit.register
    def _(self, constructor : description_logics.RoleFluentAtomicState): 
        pass
    @visit.register
    def _(self, constructor : description_logics.RoleDerivedAtomicState): 
        pass
    @visit.register
    def _(self, constructor : description_logics.RoleStaticAtomicGoal): 
        pass
    @visit.register
    def _(self, constructor : description_logics.RoleFluentAtomicGoal): 
        pass
    @visit.register
    def _(self, constructor : description_logics.RoleDerivedAtomicGoal): 
        pass
    @visit.register
    def _(self, constructor : description_logics.RoleIntersection):
        self.visit(constructor.get_left_role())
        self.visit(constructor.get_right_role())
    @visit.register
    def _(self, constructor : description_logics.RoleUnion): 
        self.visit(constructor.get_left_role())
        self.visit(constructor.get_right_role())
    @visit.register
    def _(self, constructor : description_logics.RoleComplement): 
        self.visit(constructor.get_role())
    @visit.register
    def _(self, constructor : description_logics.RoleInverse): 
        self.visit(constructor.get_role())
    @visit.register
    def _(self, constructor : description_logics.RoleComposition): 
        self.visit(constructor.get_left_role())
        self.visit(constructor.get_right_role())
    @visit.register
    def _(self, constructor : description_logics.RoleTransitiveClosure): 
        self.visit(constructor.get_role())
    @visit.register
    def _(self, constructor : description_logics.RoleReflexiveTransitiveClosure): 
        self.visit(constructor.get_role())
    @visit.register
    def _(self, constructor : description_logics.RoleRestriction): 
        self.visit(constructor.get_role())
        self.visit(constructor.get_concept())
    @visit.register
    def _(self, constructor : description_logics.RoleIdentity): 
        self.visit(constructor.get_concept())
    @visit.register
    def _(self, constructor : description_logics.BooleanStaticAtomicState): 
        pass
    @visit.register
    def _(self, constructor : description_logics.BooleanFluentAtomicState): 
        pass
    @visit.register
    def _(self, constructor : description_logics.BooleanDerivedAtomicState): 
        pass
    @visit.register
    def _(self, constructor : description_logics.BooleanConceptNonempty): 
        self.visit(constructor.get_constructor())
    @visit.register
    def _(self, constructor : description_logics.BooleanRoleNonempty): 
        self.visit(constructor.get_constructor())
    @visit.register
    def _(self, constructor : description_logics.NumericalConceptCount): 
        self.visit(constructor.get_constructor())
    @visit.register
    def _(self, constructor : description_logics.NumericalRoleCount): 
        self.visit(constructor.get_constructor())
    @visit.register
    def _(self, constructor : description_logics.NumericalDistance): 
        self.visit(constructor.get_left_concept())
        self.visit(constructor.get_role())
        self.visit(constructor.get_left_concept())


class MyCNFGrammarVisitor:
    """ Recursively visit the CNF grammar. 
        This can be useful to write a custom sentence generator.
        An available example implementation is `description_logics.GeneratorVisitor`.
    """
    def __init__(self):
        pass

    @singledispatchmethod
    def visit(self, arg):
        raise NotImplementedError("Missing implementation for the constructor!")
    
    @visit.register
    def _(self, constructor : description_logics.CNFConceptBot):
        pass
    @visit.register
    def _(self, constructor : description_logics.CNFConceptTop):
        pass
    @visit.register
    def _(self, constructor : description_logics.CNFConceptStaticAtomicState):
        pass
    @visit.register
    def _(self, constructor : description_logics.CNFConceptFluentAtomicState): 
        pass
    @visit.register
    def _(self, constructor : description_logics.CNFConceptDerivedAtomicState): 
        pass
    @visit.register
    def _(self, constructor : description_logics.CNFConceptStaticAtomicGoal): 
        pass
    @visit.register
    def _(self, constructor : description_logics.CNFConceptFluentAtomicGoal): 
        pass
    @visit.register
    def _(self, constructor : description_logics.CNFConceptDerivedAtomicGoal): 
        pass
    @visit.register
    def _(self, constructor : description_logics.CNFConceptIntersection): 
        self.visit(constructor.get_left_concept())
        self.visit(constructor.get_right_concept())
    @visit.register
    def _(self, constructor : description_logics.CNFConceptUnion): 
        self.visit(constructor.get_left_concept())
        self.visit(constructor.get_right_concept())
    @visit.register
    def _(self, constructor : description_logics.CNFConceptNegation): 
        self.visit(constructor.get_concept())
    @visit.register
    def _(self, constructor : description_logics.CNFConceptValueRestriction): 
        self.visit(constructor.get_role())
        self.visit(constructor.get_concept())
    @visit.register
    def _(self, constructor : description_logics.CNFConceptExistentialQuantification): 
        self.visit(constructor.get_role())
        self.visit(constructor.get_concept())
    @visit.register
    def _(self, constructor : description_logics.CNFConceptRoleValueMapContainment): 
        self.visit(constructor.get_left_role())
        self.visit(constructor.get_right_role())
    @visit.register
    def _(self, constructor : description_logics.CNFConceptRoleValueMapEquality): 
        self.visit(constructor.get_left_role())
        self.visit(constructor.get_right_role())
    @visit.register
    def _(self, constructor : description_logics.CNFConceptNominal): 
        pass
    @visit.register
    def _(self, constructor : description_logics.CNFRoleUniversal): 
        pass
    @visit.register
    def _(self, constructor : description_logics.CNFRoleStaticAtomicState): 
        pass
    @visit.register
    def _(self, constructor : description_logics.CNFRoleFluentAtomicState): 
        pass
    @visit.register
    def _(self, constructor : description_logics.CNFRoleDerivedAtomicState): 
        pass
    @visit.register
    def _(self, constructor : description_logics.CNFRoleStaticAtomicGoal): 
        pass
    @visit.register
    def _(self, constructor : description_logics.CNFRoleFluentAtomicGoal): 
        pass
    @visit.register
    def _(self, constructor : description_logics.CNFRoleDerivedAtomicGoal): 
        pass
    @visit.register
    def _(self, constructor : description_logics.CNFRoleIntersection):
        self.visit(constructor.get_left_role())
        self.visit(constructor.get_right_role())
    @visit.register
    def _(self, constructor : description_logics.CNFRoleUnion): 
        self.visit(constructor.get_left_role())
        self.visit(constructor.get_right_role())
    @visit.register
    def _(self, constructor : description_logics.CNFRoleComplement): 
        self.visit(constructor.get_role())
    @visit.register
    def _(self, constructor : description_logics.CNFRoleInverse): 
        self.visit(constructor.get_role())
    @visit.register
    def _(self, constructor : description_logics.CNFRoleComposition): 
        self.visit(constructor.get_left_role())
        self.visit(constructor.get_right_role())
    @visit.register
    def _(self, constructor : description_logics.CNFRoleTransitiveClosure): 
        self.visit(constructor.get_role())
    @visit.register
    def _(self, constructor : description_logics.CNFRoleReflexiveTransitiveClosure): 
        self.visit(constructor.get_role())
    @visit.register
    def _(self, constructor : description_logics.CNFRoleRestriction): 
        self.visit(constructor.get_role())
        self.visit(constructor.get_concept())
    @visit.register
    def _(self, constructor : description_logics.CNFRoleIdentity): 
        self.visit(constructor.get_concept())
    @visit.register
    def _(self, constructor : description_logics.CNFBooleanStaticAtomicState): 
        pass
    @visit.register
    def _(self, constructor : description_logics.CNFBooleanFluentAtomicState): 
        pass
    @visit.register
    def _(self, constructor : description_logics.CNFBooleanDerivedAtomicState): 
        pass
    @visit.register
    def _(self, constructor : description_logics.CNFBooleanConceptNonempty): 
        self.visit(constructor.get_constructor())
    @visit.register
    def _(self, constructor : description_logics.CNFBooleanRoleNonempty): 
        self.visit(constructor.get_constructor())
    @visit.register
    def _(self, constructor : description_logics.CNFNumericalConceptCount): 
        self.visit(constructor.get_constructor())
    @visit.register
    def _(self, constructor : description_logics.CNFNumericalRoleCount): 
        self.visit(constructor.get_constructor())
    @visit.register
    def _(self, constructor : description_logics.CNFNumericalDistance): 
        self.visit(constructor.get_left_concept())
        self.visit(constructor.get_role())
        self.visit(constructor.get_left_concept())
    @visit.register
    def _(self, nonterminal : description_logics.CNFConceptNonTerminal): 
        pass
    @visit.register
    def _(self, nonterminal : description_logics.CNFRoleNonTerminal): 
        pass
    @visit.register
    def _(self, nonterminal : description_logics.CNFBooleanNonTerminal): 
        pass
    @visit.register
    def _(self, nonterminal : description_logics.CNFNumericalNonTerminal): 
        pass
    @visit.register
    def _(self, rule : description_logics.CNFConceptDerivationRule): 
        self.visit(rule.get_head())
        self.visit(rule.get_body())
    @visit.register
    def _(self, rule : description_logics.CNFRoleDerivationRule): 
        self.visit(rule.get_head())
        self.visit(rule.get_body())
    @visit.register
    def _(self, rule : description_logics.CNFBooleanDerivationRule): 
        self.visit(rule.get_head())
        self.visit(rule.get_body())
    @visit.register
    def _(self, rule : description_logics.CNFNumericalDerivationRule): 
        self.visit(rule.get_head())
        self.visit(rule.get_body())
    @visit.register
    def _(self, rule : description_logics.CNFConceptSubstitutionRule): 
        self.visit(rule.get_head())
        self.visit(rule.get_body())
    @visit.register
    def _(self, rule : description_logics.CNFRoleSubstitutionRule): 
        self.visit(rule.get_head())
        self.visit(rule.get_body())
    @visit.register
    def _(self, rule : description_logics.CNFBooleanSubstitutionRule): 
        self.visit(rule.get_head())
        self.visit(rule.get_body())
    @visit.register
    def _(self, rule : description_logics.CNFNumericalSubstitutionRule): 
        self.visit(rule.get_head())
        self.visit(rule.get_body())
    @visit.register
    def _(self, grammar : description_logics.CNFGrammar): 
        for symbol in [grammar.get_concept_start_symbol()] + [grammar.get_role_start_symbol()] + [grammar.get_boolean_start_symbol()] + [grammar.get_numerical_start_symbol()]:
            if symbol is not None:
                self.visit(symbol)
        for rule in list(grammar.get_concept_derivation_rules()) + list(grammar.get_role_derivation_rules()) + list(grammar.get_boolean_derivation_rules()) + list(grammar.get_numerical_derivation_rules()):
            self.visit(rule)
        for rule in list(grammar.get_concept_substitution_rules()) + list(grammar.get_role_substitution_rules()) + list(grammar.get_boolean_substitution_rules()) + list(grammar.get_numerical_substitution_rules()):
            self.visit(rule)


def main():
    domain_filepath = str(ROOT_DIR / "data" / "gripper" / "domain.pddl")
    problem_filepath = str(ROOT_DIR / "data" / "gripper" / "p-1-0.pddl")

    # Create GeneralizedSearchContext
    search_context_options = search.SearchContextOptions()
    search_context_options.mode = search.SearchMode.GROUNDED
    generalized_search_context = search.GeneralizedSearchContext.create(domain_filepath, [problem_filepath], search_context_options)

    # Create KnowledgeBase
    state_space_options = datasets.StateSpaceOptions()
    state_space_options.symmetry_pruning = True 
    generalized_state_space_options = datasets.GeneralizedStateSpaceOptions()
    knowledge_base_options = datasets.KnowledgeBaseOptions()
    knowledge_base_options.state_space_options = state_space_options
    knowledge_base_options.generalized_state_space_options = generalized_state_space_options
    knowledge_base = datasets.KnowledgeBase.create(generalized_search_context, knowledge_base_options)

    # Create a CNFGrammar for the domain.
    # You can also write down your own grammar in a BNF specification that will automatically be translated into a CNF representation.
    grammar = description_logics.CNFGrammar.create(description_logics.GrammarSpecificationEnum.FRANCES_ET_AL_AAAI2021, knowledge_base.get_domain())

    denotation_repositories = description_logics.DenotationRepositories()
    pruning_function = description_logics.StateListRefinementPruningFunction(knowledge_base.get_generalized_state_space(), denotation_repositories)
    sentences = description_logics.GeneratedSentencesContainer()
    repositories = description_logics.Repositories()
    sentence_generator = description_logics.GeneratorVisitor(pruning_function, sentences, repositories, 5)
    sentence_generator.visit(grammar)

    print("================================================================================")
    print("Input Grammar in CNF")
    print("================================================================================")
    print(grammar)
    # Visit the grammar
    MyCNFGrammarVisitor().visit(grammar)

    print("================================================================================")
    print("Generated concepts")
    print("================================================================================")
    for complexity, constructors in enumerate(sentences.get_concepts().get(grammar.get_concept_start_symbol())):
        print("Complexity:", complexity)
        for constructor in constructors:
            # Print the sentence
            print(f"    {constructor}")
            # Check whether the sentence matches grammar specification
            assert(grammar.test_match(constructor))
            # Visit the constructor
            MyConstructorVisitor().visit(constructor)

    print("================================================================================")
    print("Generated roles")
    print("================================================================================")
    for complexity, constructors in enumerate(sentences.get_roles().get(grammar.get_role_start_symbol())):
        print("Complexity:", complexity)
        for constructor in constructors:
            # Print the sentence
            print(f"    {constructor}")
            # Check whether the sentence matches grammar specification
            assert(grammar.test_match(constructor))
            # Visit the constructor
            MyConstructorVisitor().visit(constructor)

    print("================================================================================")
    print("Generated booleans")
    print("================================================================================")
    for complexity, constructors in enumerate(sentences.get_booleans().get(grammar.get_boolean_start_symbol())):
        print("Complexity:", complexity)
        for constructor in constructors:
            # Print the sentence
            print(f"    {constructor}")
            # Check whether the sentence matches grammar specification
            assert(grammar.test_match(constructor))
            # Visit the constructor
            MyConstructorVisitor().visit(constructor)

    print("================================================================================")
    print("Generated numericals")
    print("================================================================================")
    for complexity, constructors in enumerate(sentences.get_numericals().get(grammar.get_numerical_start_symbol())):
        print("Complexity:", complexity)
        for constructor in constructors:
            # Print the sentence
            print(f"    {constructor}")
            # Check whether the sentence matches grammar specification
            assert(grammar.test_match(constructor))
            # Visit the constructor
            MyConstructorVisitor().visit(constructor)

    print("================================================================================")
    print("Evaluate numericals")
    print("================================================================================")
    
    for complexity, constructors in enumerate(sentences.get_numericals().get(grammar.get_numerical_start_symbol())):
        print("Complexity:", complexity)
        for constructor in constructors:
            print(f"    {constructor}")
            for state_space in knowledge_base.get_state_spaces():
                for vertex in state_space.get_graph().get_vertices():
                    state = datasets.get_state(vertex)
                    problem = datasets.get_problem(vertex)
                    context = description_logics.EvaluationContext(state, problem, denotation_repositories)
                    value = constructor.evaluate(context).get_data()
                    print("State:", state.to_string(problem))
                    print("Value:", value)
    
    

if __name__ == "__main__":
    main()



# Import all classes for better IDE support


from pymimir.pymimir.advanced.graphs import (
    # Color and graph isomorphism (nauty)
    Color,
    NautySparseGraph,

    # Vertices
    PyVertex,
    EmptyVertex,
    ColoredVertex,

    # Edges
    PyEdge,
    EmptyEdge,
    ColoredEdge,

    # StaticPyGraph
    StaticPyGraph,
    ImmutableStaticPyGraph,
    ForwardStaticPyGraph,
    BidirectionalStaticPyGraph,
    # DynamicPyGraph
    DynamicPyGraph,

    # StaticEmptyGraph
    StaticEmptyGraph,
    ImmutableStaticEmptyGraph,
    ForwardStaticEmptyGraph,
    BidirectionalStaticEmptyGraph,

    # StaticVertexColoredGraph
    StaticVertexColoredGraph,
    ImmutableStaticVertexColoredGraph,
    ForwardStaticVertexColoredGraph,
    BidirectionalStaticVertexColoredGraph,

    # StaticEdgeColoredGraph
    StaticEdgeColoredGraph,
    ImmutableStaticEdgeColoredGraph,
    ForwardStaticEdgeColoredGraph,
    BidirectionalStaticEdgeColoredGraph,
)

# Coloring
from pymimir.pymimir.advanced.graphs import (
    compute_color_refinement_certificate,
    compute_2fwl_certificate,
    compute_3fwl_certificate,
    compute_4fwl_certificate,
    CertificateColorRefinement,
    KFWLIsomorphismTypeCompressionFunction,
    Certficate2FWL,
    Certficate3FWL,
    Certficate4FWL,
)

# graph properties
from pymimir.pymimir.advanced.graphs import (
    is_undirected,
    is_loopless,
    is_acyclic,
    is_multi,
)



from functools import lru_cache
from pathlib import Path
from typing import TypeAlias, Union

from pymimir.advanced.formalism import Action as AdvancedAction
from pymimir.advanced.formalism import ConditionalEffect as AdvancedConditionalEffect
from pymimir.advanced.formalism import ConjunctiveCondition as AdvancedConjunctiveCondition
from pymimir.advanced.formalism import GroundConjunctiveCondition as AdvancedGroundConjunctiveCondition
from pymimir.advanced.formalism import ConjunctiveEffect as AdvancedConjunctiveEffect
from pymimir.advanced.formalism import DerivedAtom as AdvancedDerivedAtom
from pymimir.advanced.formalism import DerivedGroundAtom as AdvancedDerivedGroundAtom
from pymimir.advanced.formalism import DerivedGroundLiteral as AdvancedDerivedGroundLiteral
from pymimir.advanced.formalism import DerivedLiteral as AdvancedDerivedLiteral
from pymimir.advanced.formalism import DerivedLiteralList as AdvancedDerivedLiteralList
from pymimir.advanced.formalism import DerivedPredicate as AdvancedDerivedPredicate
from pymimir.advanced.formalism import Domain as AdvancedDomain
from pymimir.advanced.formalism import FluentAtom as AdvancedFluentAtom
from pymimir.advanced.formalism import FluentGroundAtom as AdvancedFluentGroundAtom
from pymimir.advanced.formalism import FluentGroundLiteral as AdvancedFluentGroundLiteral
from pymimir.advanced.formalism import FluentLiteral as AdvancedFluentLiteral
from pymimir.advanced.formalism import FluentLiteralList as AdvancedFluentLiteralList
from pymimir.advanced.formalism import FluentPredicate as AdvancedFluentPredicate
# from pymimir.advanced.formalism import NumericConstraintList as AdvancedNumericConstraintList  # TODO: Dominik -- Cannot import anymore.
from pymimir.advanced.formalism import Object as AdvancedObject
from pymimir.advanced.formalism import ObjectList as AdvancedObjectList
from pymimir.advanced.formalism import Parser as AdvancedParser
from pymimir.advanced.formalism import Problem as AdvancedProblem
from pymimir.advanced.formalism import Repositories as AdvancedRepositories
from pymimir.advanced.formalism import StaticAtom as AdvancedStaticAtom
from pymimir.advanced.formalism import StaticGroundAtom as AdvancedStaticGroundAtom
from pymimir.advanced.formalism import StaticGroundLiteral as AdvancedStaticGroundLiteral
from pymimir.advanced.formalism import StaticLiteral as AdvancedStaticLiteral
from pymimir.advanced.formalism import StaticLiteralList as AdvancedStaticLiteralList
from pymimir.advanced.formalism import StaticPredicate as AdvancedStaticPredicate
from pymimir.advanced.formalism import Term as AdvancedTerm
from pymimir.advanced.formalism import TermList as AdvancedTermList
from pymimir.advanced.formalism import Variable as AdvancedVariable
from pymimir.advanced.formalism import VariableList as AdvancedVariableList
from pymimir.advanced.search import GroundedAxiomEvaluator as AdvancedGroundedAxiomEvaluator
from pymimir.advanced.search import LiftedAxiomEvaluator as AdvancedLiftedAxiomEvaluator
from pymimir.advanced.search import SearchMode, SearchContext, SearchContextOptions
from pymimir.advanced.search import State as AdvancedState
from pymimir.advanced.search import LiftedApplicableActionGenerator as AdvancedLiftedApplicableActionGenerator
# from pymimir.advanced.formalism import AuxiliaryGroundNumericEffect as AdvancedAuxiliaryGroundNumericEffect
# from pymimir.advanced.formalism import AuxiliaryNumericEffect as AdvancedAuxiliaryNumericEffect
# from pymimir.advanced.formalism import FluentGroundNumericEffect as AdvancedFluentGroundNumericEffect
# from pymimir.advanced.formalism import FluentNumericEffect as AdvancedFluentNumericEffect
# from pymimir.advanced.formalism import FunctionExpression as AdvancedFunctionExpression
# from pymimir.advanced.formalism import NumericConstraint as AdvancedNumericConstraint


AdvancedAtom: TypeAlias = Union[AdvancedStaticAtom, AdvancedFluentAtom, AdvancedDerivedAtom]
AdvancedGroundAtom: TypeAlias = Union[AdvancedStaticGroundAtom, AdvancedFluentGroundAtom, AdvancedDerivedGroundAtom]
AdvancedLiteral: TypeAlias = Union[AdvancedStaticLiteral, AdvancedFluentLiteral, AdvancedDerivedLiteral]
AdvancedGroundLiteral: TypeAlias = Union[AdvancedStaticGroundLiteral, AdvancedFluentGroundLiteral, AdvancedDerivedGroundLiteral]
AdvancedPredicate: TypeAlias = Union[AdvancedStaticPredicate, AdvancedFluentPredicate, AdvancedDerivedPredicate]
AdvancedAxiomEvaluator: TypeAlias = Union[AdvancedGroundedAxiomEvaluator, AdvancedLiftedAxiomEvaluator]
# AdvancedNumericEffect: TypeAlias = Union[AdvancedFluentNumericEffect, AdvancedAuxiliaryNumericEffect]
# AdvancedGroundNumericEffect: TypeAlias = Union[AdvancedFluentGroundNumericEffect, AdvancedAuxiliaryGroundNumericEffect]


class Variable:
    _advanced_variable: AdvancedVariable = None

    def __init__(self, advanced_variable: AdvancedVariable) -> None:
        assert isinstance(advanced_variable, AdvancedVariable), "Invalid variable type."
        self._advanced_variable = advanced_variable

    def get_index(self) -> 'int':
        """Get the index of the variable."""
        return self._advanced_variable.get_index()

    def get_name(self) -> 'str':
        """Get the name of the variable."""
        return self._advanced_variable.get_name()

    def __str__(self) -> 'str':
        """Get the string representation of the variable."""
        return str(self._advanced_variable)

    def __repr__(self) -> 'str':
        """Get the string representation of the variable."""
        return repr(self._advanced_variable)

    def __hash__(self) -> 'int':
        """Get the hash of the variable."""
        return hash(self._advanced_variable)

    def __eq__(self, other: object) -> bool:
        """Check if two variables are equal."""
        if not isinstance(other, Variable):
            return False
        return self._advanced_variable == other._advanced_variable


class Object:
    """Object class for the PDDL domain."""
    _advanced_object: AdvancedObject = None

    def __init__(self, advanced_object: AdvancedObject) -> None:
        assert isinstance(advanced_object, AdvancedObject), "Invalid object type."
        self._advanced_object = advanced_object

    def get_index(self) -> 'int':
        """Get the index of the object."""
        return self._advanced_object.get_index()

    def get_name(self) -> 'str':
        """Get the name of the object."""
        return self._advanced_object.get_name()

    def __str__(self) -> 'str':
        """Get the string representation of the object."""
        return str(self._advanced_object)

    def __repr__(self) -> 'str':
        """Get the string representation of the object."""
        return repr(self._advanced_object)

    def __hash__(self) -> 'int':
        """Get the hash of the object."""
        return hash(self._advanced_object)

    def __eq__(self, other: object) -> bool:
        """Check if two objects are equal."""
        if not isinstance(other, Object):
            return False
        return self._advanced_object == other._advanced_object


Term: TypeAlias = Union[Object, Variable]


class Predicate:
    """Predicate class for the PDDL domain."""
    _advanced_predicate: AdvancedPredicate = None

    def __init__(self, advanced_predicate: AdvancedPredicate) -> None:
        assert isinstance(advanced_predicate, AdvancedPredicate), "Invalid predicate type."
        self._advanced_predicate = advanced_predicate

    def get_index(self) -> 'int':
        """Get the index of the predicate."""
        return self._advanced_predicate.get_index()

    def get_name(self) -> 'str':
        """Get the name of the predicate."""
        return self._advanced_predicate.get_name()

    def get_arity(self) -> 'int':
        """Get the arity of the predicate."""
        return self._advanced_predicate.get_arity()

    def get_parameters(self) -> 'list[Variable]':
        """Get the parameters of the predicate."""
        return [Variable(x) for x in self._advanced_predicate.get_parameters()]

    def __str__(self):
        """Get the string representation of the predicate."""
        return str(self._advanced_predicate)

    def __repr__(self):
        """Get the string representation of the predicate."""
        return repr(self._advanced_predicate)

    def __hash__(self) -> 'int':
        """Get the hash of the predicate."""
        return hash(self._advanced_predicate)

    def __eq__(self, other: object) -> bool:
        """Check if two predicates are equal."""
        if not isinstance(other, Predicate):
            return False
        return self._advanced_predicate == other._advanced_predicate


class GroundAtom:
    """GroundAtom class for the PDDL domain."""
    _advanced_ground_atom: AdvancedGroundAtom = None

    def __init__(self, advanced_ground_atom: AdvancedGroundAtom) -> None:
        assert isinstance(advanced_ground_atom, AdvancedGroundAtom), "Invalid ground atom type."
        self._advanced_ground_atom = advanced_ground_atom

    def get_index(self) -> 'int':
        """Get the index of the ground atom."""
        return self._advanced_ground_atom.get_index()

    def get_predicate(self) -> 'Predicate':
        """Get the predicate of the ground atom."""
        return Predicate(self._advanced_ground_atom.get_predicate())

    def get_terms(self) -> 'list[Object]':
        """Get the objects of the ground atom."""
        return [Object(x) for x in self._advanced_ground_atom.get_objects()]

    def get_arity(self) -> 'int':
        """Get the arity of the ground atom."""
        return self._advanced_ground_atom.get_arity()

    def __str__(self):
        """Get the string representation of the ground atom."""
        return str(self._advanced_ground_atom)

    def __repr__(self):
        """Get the string representation of the ground atom."""
        return repr(self._advanced_ground_atom)

    def __hash__(self) -> 'int':
        """Get the hash of the ground atom."""
        return hash(self._advanced_ground_atom)

    def __eq__(self, other: object) -> bool:
        """Check if two ground atoms are equal."""
        if not isinstance(other, GroundAtom):
            return False
        return self._advanced_ground_atom == other._advanced_ground_atom


class Atom:
    """Atom class for the PDDL domain."""
    _advanced_atom: AdvancedAtom = None

    def __init__(self, advanced_atom: AdvancedAtom) -> None:
        assert isinstance(advanced_atom, AdvancedAtom), "Invalid atom type."
        self._advanced_atom = advanced_atom

    def get_index(self) -> 'int':
        """Get the index of the atom."""
        return self._advanced_atom.get_index()

    def get_predicate(self) -> 'Predicate':
        """Get the predicate of the atom."""
        return Predicate(self._advanced_atom.get_predicate())

    def get_terms(self) -> 'list[Union[Object, Variable]]':
        """Get the terms of the atom."""
        return [Object(x) if isinstance(x, AdvancedObject) else Variable(x) for x in self._advanced_atom.get_terms()]

    def get_variables(self) -> 'list[Variable]':
        """Get the variables of the atom."""
        return [Variable(x) for x in self._advanced_atom.get_variables()]

    def get_arity(self) -> 'int':
        """Get the arity of the atom."""
        return len(self.get_terms())

    def __str__(self):
        """Get the string representation of the atom."""
        return str(self._advanced_atom)

    def __repr__(self):
        """Get the string representation of the atom."""
        return repr(self._advanced_atom)

    def __hash__(self) -> 'int':
        """Get the hash of the atom."""
        return hash(self._advanced_atom)

    def __eq__(self, other: object) -> bool:
        """Check if two atoms are equal."""
        if not isinstance(other, Atom):
            return False
        return self._advanced_atom == other._advanced_atom


class GroundLiteral:
    """GroundLiteral class for the PDDL domain."""
    _advanced_ground_literal: AdvancedGroundLiteral = None

    def __init__(self, advanced_ground_literal: AdvancedGroundLiteral) -> None:
        assert isinstance(advanced_ground_literal, AdvancedGroundLiteral), "Invalid ground literal type."
        self._advanced_ground_literal = advanced_ground_literal

    def get_index(self) -> 'int':
        """Get the index of the ground literal."""
        return self._advanced_ground_literal.get_index()

    def get_atom(self) -> 'GroundAtom':
        """Get the atom of the ground literal."""
        return GroundAtom(self._advanced_ground_literal.get_atom())

    def get_polarity(self) -> bool:
        """Get the polarity of the ground literal."""
        return self._advanced_ground_literal.get_polarity()

    def __str__(self):
        """Get the string representation of the ground literal."""
        return str(self._advanced_ground_literal)

    def __repr__(self):
        """Get the string representation of the ground literal."""
        return repr(self._advanced_ground_literal)

    def __hash__(self) -> 'int':
        """Get the hash of the ground literal."""
        return hash(self._advanced_ground_literal)

    def __eq__(self, other: object) -> bool:
        """Check if two ground literals are equal."""
        if not isinstance(other, GroundLiteral):
            return False
        return self._advanced_ground_literal == other._advanced_ground_literal


class Literal:
    """Literal class for the PDDL domain."""
    _advanced_literal = None

    def __init__(self, advanced_literal: AdvancedLiteral) -> None:
        assert isinstance(advanced_literal, AdvancedLiteral), "Invalid literal type."
        self._advanced_literal = advanced_literal

    def get_index(self) -> 'int':
        """Get the index of the literal."""
        return self._advanced_literal.get_index()

    def get_atom(self) -> 'Atom':
        """Get the atom of the literal."""
        return Atom(self._advanced_literal.get_atom())

    def get_polarity(self) -> bool:
        """Get the polarity of the literal."""
        return self._advanced_literal.get_polarity()

    def __str__(self) -> 'str':
        """Get the string representation of the literal."""
        return str(self._advanced_literal)

    def __repr__(self) -> 'str':
        """Get the string representation of the literal."""
        return repr(self._advanced_literal)

    def __hash__(self) -> 'int':
        """Get the hash of the literal."""
        return hash(self._advanced_literal)

    def __eq__(self, other: object) -> bool:
        """Check if two literals are equal."""
        if not isinstance(other, Literal):
            return False
        return self._advanced_literal == other._advanced_literal


def _split_literal_list(literals: list[Literal]) -> 'tuple[AdvancedStaticLiteralList, AdvancedFluentLiteralList, AdvancedDerivedLiteralList]':
    """Split the literal list into static, fluent, and derived literals."""
    advanced_static_literals = AdvancedStaticLiteralList([x._advanced_literal for x in literals if isinstance(x, AdvancedStaticLiteral)])
    advanced_fluent_literals = AdvancedFluentLiteralList([x._advanced_literal for x in literals if isinstance(x, AdvancedFluentLiteral)])
    advanced_derived_literals = AdvancedDerivedLiteralList([x._advanced_literal for x in literals if isinstance(x, AdvancedDerivedLiteral)])
    return advanced_static_literals, advanced_fluent_literals, advanced_derived_literals
    #

# class FunctionExpression:
#     """FunctionExpression class for the PDDL domain."""
#     _advanced_function_expression: AdvancedFunctionExpression = None

#     def __init__(self, advanced_function_expression: AdvancedFunctionExpression) -> None:
#         self._advanced_function_expression = advanced_function_expression
#         raise NotImplementedError("Not implemented.")


# class NumericConstraint:
#     """NumericConstraint class for the PDDL domain."""
#     _advanced_numeric_constraint: AdvancedNumericConstraint = None

#     def __init__(self, advanced_numeric_constraint: AdvancedNumericConstraint) -> None:
#         self._advanced_numeric_constraint = advanced_numeric_constraint

#     def get_index(self) -> 'int':
#         """Get the index of the numeric constraint."""
#         return self._advanced_numeric_constraint.get_index()

#     def get_binary_comparator(self):
#         """Get the binary comparator of the numeric constraint."""
#         raise NotImplementedError("Not implemented.")
#         return self._advanced_numeric_constraint.get_binary_comparator()

#     def get_left_function_expression(self) -> 'FunctionExpression':
#         """Get the left function expression of the numeric constraint."""
#         return self._advanced_numeric_constraint.get_left_function_expression()

#     def get_right_function_expression(self) -> 'FunctionExpression':
#         """Get the right function expression of the numeric constraint."""
#         return self._advanced_numeric_constraint.get_right_function_expression()


# class NumericEffect:
#     """NumericEffect class for the PDDL domain."""
#     _advanced_numeric_effect: AdvancedNumericEffect = None

#     def __init__(self, advanced_numeric_effect: AdvancedNumericEffect) -> None:
#         self._advanced_numeric_effect = advanced_numeric_effect

#     def get_index(self) -> 'int':
#         """Get the index of the numeric effect."""
#         return self._advanced_numeric_effect.get_index()

#     def get_function_expression(self) -> 'FunctionExpression':
#         """Get the function expression of the numeric effect."""
#         return self._advanced_numeric_effect.get_function_expression()

#     def get_function(self) -> 'FunctionExpression':
#         """Get the function of the numeric effect."""
#         raise NotImplementedError("Not implemented.")


class ConjunctiveEffect:
    """ConjunctiveEffect class for the PDDL domain."""
    _advanced_conjunctive_effect: AdvancedConjunctiveEffect = None

    def __init__(self, advanced_effect: AdvancedConjunctiveEffect) -> None:
        assert isinstance(advanced_effect, AdvancedConjunctiveEffect), "Invalid conjunctive effect type."
        self._advanced_conjunctive_effect = advanced_effect

    def get_index(self) -> 'int':
        """Get the index of the conjunctive effect."""
        return self._advanced_conjunctive_effect.get_index()

    def get_parameters(self) -> 'list[Variable]':
        """Get the parameters of the conjunctive effect."""
        return [Variable(x) for x in self._advanced_conjunctive_effect.get_parameters()]

    def get_literals(self) -> 'list[Literal]':
        """Get the literals of the conjunctive effect."""
        return [Literal(x) for x in self._advanced_conjunctive_effect.get_literals()]

    # def get_numeric_effects(self) -> 'list[NumericEffect]':
    #     """Get the fluent numeric effects of the conjunctive effect."""
    #     numeric_effects = list(self._advanced_conjunctive_effect.get_fluent_numeric_effects())
    #     auxiliary_numeric_effect = self._advanced_conjunctive_effect.get_auxiliary_numeric_effect()
    #     if auxiliary_numeric_effect: numeric_effects.append(auxiliary_numeric_effect)
    #     return numeric_effects

    def __str__(self):
        """Get the string representation of the conjunctive effect."""
        return str(self._advanced_conjunctive_effect)

    def __repr__(self):
        """Get the string representation of the conjunctive effect."""
        return repr(self._advanced_conjunctive_effect)

    def __hash__(self) -> 'int':
        """Get the hash of the conjunctive effect."""
        return hash(self._advanced_conjunctive_effect)

    def __eq__(self, other: object) -> bool:
        """Check if two conjunctive effects are equal."""
        if not isinstance(other, ConjunctiveEffect):
            return False
        return self._advanced_conjunctive_effect == other._advanced_conjunctive_effect


class ConditionalEffect:
    """ConditionalEffect class for the PDDL domain."""
    _advanced_conditional_effect: AdvancedConditionalEffect = None

    def __init__(self, advanced_conditional_effect: AdvancedConditionalEffect) -> None:
        assert isinstance(advanced_conditional_effect, AdvancedConditionalEffect), "Invalid conditional effect type."
        self._advanced_conditional_effect = advanced_conditional_effect

    def get_index(self) -> 'int':
        """Get the index of the conditional effect."""
        return self._advanced_conditional_effect.get_index()

    def get_parameters(self) -> 'ConjunctiveCondition':
        """Get the parameters of the conditional effect."""
        return ConjunctiveCondition(self._advanced_conditional_effect.get_parameters())

    def get_static_conditions(self) -> 'ConjunctiveEffect':
        """Get the static conditions of the conditional effect."""
        return ConjunctiveEffect(self._advanced_conditional_effect.get_static_conditions())

    def __str__(self):
        """Get the string representation of the conditional effect."""
        return str(self._advanced_conditional_effect)

    def __repr__(self):
        """Get the string representation of the conditional effect."""
        return repr(self._advanced_conditional_effect)

    def __hash__(self) -> 'int':
        """Get the hash of the conditional effect."""
        return hash(self._advanced_conditional_effect)

    def __eq__(self, other: object) -> bool:
        """Check if two conditional effects are equal."""
        if not isinstance(other, ConditionalEffect):
            return False
        return self._advanced_conditional_effect == other._advanced_conditional_effect


class Action:
    _advanced_action: AdvancedAction = None

    def __init__(self, advanced_action: AdvancedAction) -> None:
        assert isinstance(advanced_action, AdvancedAction), "Invalid action type."
        self._advanced_action = advanced_action

    def get_index(self) -> 'int':
        """Get the index of the action."""
        return self._advanced_action.get_index()

    def get_arity(self) -> 'int':
        """Get the arity of the action."""
        return self._advanced_action.get_arity()

    def get_name(self) -> 'str':
        """Get the name of the action."""
        return self._advanced_action.get_name()

    def get_parameters(self) -> 'list[Variable]':
        """Get the parameters of the action."""
        return [Variable(x) for x in self._advanced_action.get_parameters()]

    def get_conjunctive_condition(self) -> 'ConjunctiveCondition':
        """Get the conjunctive condition of the action."""
        return ConjunctiveCondition(self._advanced_action.get_conjunctive_condition())

    def get_conjunctive_effect(self) -> 'ConjunctiveEffect':
        """Get the conjunctive effect of the action."""
        return ConjunctiveEffect(self._advanced_action.get_conjunctive_effect())

    def get_conditional_effect(self) -> 'list[ConditionalEffect]':
        """Get the conditional effect of the action."""
        return [ConditionalEffect(x) for x in self._advanced_action.get_conditional_effects()]

    def __str__(self):
        """Get the string representation of the action."""
        return str(self._advanced_action)

    def __repr__(self):
        """Get the string representation of the action."""
        return repr(self._advanced_action)

    def __hash__(self) -> 'int':
        """Get the hash of the action."""
        return hash(self._advanced_action)

    def __eq__(self, other: object) -> bool:
        """Check if two actions are equal."""
        if not isinstance(other, Action):
            return False
        return self._advanced_action == other._advanced_action


class Domain:
    """Domain class for the PDDL domain."""
    _advanced_parser: AdvancedParser = None
    _advanced_domain: AdvancedDomain = None
    _repositories: AdvancedRepositories = None

    def __init__(self, domain_path: Union[Path, str]) -> None:
        """
        Initialize the domain with the given path.

        Args:
            domain_path (Path): The file path to the domain definition file.
        """
        assert isinstance(domain_path, (Path, str)), "Invalid domain path type."
        self._advanced_parser = AdvancedParser(str(domain_path))
        self._advanced_domain = self._advanced_parser.get_domain()
        self._repositories = self._advanced_domain.get_repositories()

    def get_name(self) -> 'str':
        return self._advanced_domain.get_name()

    def get_requirements(self) -> 'list[str]':
        """Get the requirements of the domain."""
        requirements = str(self._advanced_domain.get_requirements()).removeprefix('(:requirements').removesuffix(')').split()
        return requirements

    def get_actions(self) -> 'list[Action]':
        """Get the actions of the domain."""
        return [Action(x) for x in self._advanced_domain.get_actions()]

    def get_action(self, action_name: str) -> 'Action':
        """Get the action of the domain."""
        for action in self.get_actions():
            if action.get_name() == action_name:
                return action
        raise ValueError(f"Action '{action_name}' not found in the domain.")

    def get_constants(self) -> 'list[Object]':
        """Get the constants of the domain."""
        return [Object(x) for x in self._advanced_domain.get_constants()]

    def get_constant(self, constant_name: str) -> 'Object':
        """Get the constant of the domain."""
        for constant in self.get_constants():
            if constant.get_name() == constant_name:
                return constant
        raise ValueError(f"Constant '{constant_name}' not found in the domain.")

    def get_predicates(self, skip_static = False, skip_fluent = False, skip_derived = False) -> 'list[Predicate]':
        """Get the predicates of the domain."""
        predicates = []
        if not skip_static:
            predicates.extend([Predicate(x) for x in self._advanced_domain.get_static_predicates()])
        if not skip_fluent:
            predicates.extend([Predicate(x) for x in self._advanced_domain.get_fluent_predicates()])
        if not skip_derived:
            predicates.extend([Predicate(x) for x in self._advanced_domain.get_derived_predicates()])
        return predicates

    def get_predicate(self, predicate_name: str) -> 'Predicate':
        """Get the predicate of the domain."""
        for predicate in self.get_predicates():
            if predicate.get_name() == predicate_name:
                return predicate
        raise ValueError(f"Predicate '{predicate_name}' not found in the domain.")


class Problem:
    """Problem class for the PDDL problem."""
    _domain: Domain = None
    _search_context: SearchContext = None
    _advanced_problem: AdvancedProblem = None

    def __init__(self, domain: Domain, problem_path: Union[Path, str], mode: str = 'lifted') -> None:
        """
        Initialize the problem with the given domain, path, and mode.

        Args:
            domain (Domain): The PDDL domain.
            problem_path (Path): The path to the problem definition file.
            mode (str): The search mode, either 'lifted' or 'grounded'.
        """
        assert isinstance(domain, Domain), "Invalid domain type."
        assert isinstance(problem_path, (Path, str)), "Invalid problem path type."
        if mode not in ['lifted', 'grounded']:
            raise ValueError("Invalid mode. Use 'lifted' or 'grounded'.")
        search_mode = SearchMode.LIFTED if mode == 'lifted' else SearchMode.GROUNDED
        self._domain = domain
        advanced_problem = domain._advanced_parser.parse_problem(str(problem_path))
        self._search_context = SearchContext.create(advanced_problem, SearchContextOptions(search_mode))
        self._advanced_problem = self._search_context.get_problem()

    def _to_advanced_term(self, term: Term) -> 'AdvancedTerm':
        assert isinstance(term, (Object, Variable)), "Invalid term type."
        advanced_term = term._advanced_object if isinstance(term, Object) else term._advanced_variable
        return self._advanced_problem.get_or_create_term(advanced_term)

    def get_index(self) -> 'int':
        """Get the index of the problem."""
        return self._advanced_problem.get_index()

    def get_name(self) -> 'str':
        """Get the name of the problem."""
        return self._advanced_problem.get_name()

    def get_objects(self) -> 'list[Object]':
        """Get the objects of the problem."""
        return [Object(x) for x in self._advanced_problem.get_objects()]

    def get_object(self, object_name: str) -> 'Object':
        """Get the object of the problem."""
        for obj in self.get_objects():
            if obj.get_name() == object_name:
                return obj
        raise ValueError(f"Object '{object_name}' not found in the problem.")

    def get_requirements(self) -> 'list[str]':
        """Get the requirements of the problem."""
        requirements = str(self._advanced_problem.get_requirements()).removeprefix('(:requirements').removesuffix(')').split()
        return requirements

    def get_initial_state(self) -> 'State':
        """Get the initial state of the problem."""
        initial_state, _ = self._search_context.get_state_repository().get_or_create_initial_state()
        return State(self, initial_state)

    def get_initial_atoms(self, skip_static = False, skip_fluent = False, skip_derived = False) -> 'list[GroundAtom]':
        """Get the initial atoms of the problem."""
        problem: AdvancedProblem = self._search_context.get_problem()
        initial_atoms = []
        # Add static atoms if skip_static is False
        if not skip_static:
            static_atoms = problem.get_static_initial_atoms()
            initial_atoms.extend([GroundAtom(x) for x in static_atoms])
        # Add fluent atoms
        if not skip_fluent:
            fluent_atoms = problem.get_fluent_initial_atoms()
            initial_atoms.extend([GroundAtom(x) for x in fluent_atoms])
        # Add derived atoms if skip_derived is False
        if not skip_derived:
            initial_state, _ = self._search_context.get_state_repository().get_or_create_initial_state()
            derived_indices = initial_state.get_derived_atoms()
            derived_atoms = problem.get_repositories().get_derived_ground_atoms_from_indices(derived_indices)
            initial_atoms.extend([GroundAtom(x) for x in derived_atoms])
        return initial_atoms

    def get_goal_condition(self) -> 'GroundConjunctiveCondition':
        """Get the goal condition of the problem."""
        static_goal = self._advanced_problem.get_static_goal_condition()
        fluent_goal = self._advanced_problem.get_fluent_goal_condition()
        derived_goal = self._advanced_problem.get_derived_goal_condition()
        return GroundConjunctiveCondition()  # TODO: Dominik -- How do I create one?

    def new_atom(self, predicate: Predicate, terms: list[Term]) -> 'Atom':
        """Create an atom from the given predicate and terms."""
        assert isinstance(predicate, Predicate), "Invalid predicate type."
        advanced_terms = AdvancedTermList([self._to_advanced_term(term) for term in terms])
        return Atom(self._advanced_problem.get_or_create_atom(predicate._advanced_predicate, advanced_terms))

    def new_literal(self, atom: Atom, polarity: bool) -> 'Literal':
        """Create a literal with the given polarity."""
        assert isinstance(atom, Atom), "Invalid atom type."
        return Literal(self._advanced_problem.get_or_create_literal(polarity, atom._advanced_atom))

    def new_ground_atom(self, predicate: Predicate, objects: list[Object]) -> 'GroundAtom':
        """Create a ground atom from the given predicate and objects."""
        assert isinstance(predicate, Predicate), "Invalid predicate type."
        advanced_objects = AdvancedObjectList([x._advanced_object for x in objects])
        return GroundAtom(self._advanced_problem.get_or_create_ground_atom(predicate._advanced_predicate, advanced_objects))

    def new_ground_literal(self, ground_atom: GroundAtom, polarity: bool) -> 'GroundLiteral':
        """Create a ground literal with the given polarity."""
        assert isinstance(ground_atom, GroundAtom), "Invalid ground atom type."
        predicate = ground_atom.get_predicate()
        # variables = predicate.get_parameters()  # TODO: This does not work, for some reason.
        variables = self.new_variable_list(['?x' + str(i + 1) for i in range(predicate.get_arity())])  # TODO: Why does this resolve the issue?
        advanced_terms = AdvancedTermList([self._to_advanced_term(variable) for variable in variables])
        advanced_atom = self._advanced_problem.get_or_create_atom(predicate._advanced_predicate, advanced_terms)
        advanced_literal = self._advanced_problem.get_or_create_literal(polarity, advanced_atom)
        advanced_objects = ground_atom._advanced_ground_atom.get_objects()
        return self._advanced_problem.ground(advanced_literal, advanced_objects)

    def new_variable(self, name: str, parameter_index: int) -> 'Variable':
        return Variable(self._advanced_problem.get_or_create_variable(name, parameter_index))

    def new_variable_list(self, names: list[str]) -> 'list[Variable]':
        return [self.new_variable(name, parameter_index) for name, parameter_index in zip(names, range(len(names)))]

    def new_conjunctive_condition(self, variables: list[Variable], literals: list[Literal]) -> 'ConjunctiveCondition':
        """Create a conjunctive condition from the given parameters, literals, and ground literals."""
        advanced_variables = AdvancedVariableList([x._advanced_variable for x in variables])
        advanced_static_literals, advanced_fluent_literals, advanced_derived_literals = _split_literal_list(literals)
        advanced_numeric_constraints = AdvancedNumericConstraintList()
        return ConjunctiveCondition(self._advanced_problem.get_or_create_conjunctive_condition(advanced_variables,
                                                                                               advanced_static_literals,
                                                                                               advanced_fluent_literals,
                                                                                               advanced_derived_literals,
                                                                                               advanced_numeric_constraints))
        # TODO: Dominik -- Python binding needs to be fixed because of Hana bindings.


class State:
    """
    State class representing a node in the search space.
    """
    def __init__(self, problem: Problem, advanced_state: AdvancedState):
        self._problem = problem
        self._advanced_state = advanced_state

    def get_index(self) -> 'int':
        """
        Returns the index of the state.
        """
        return self._advanced_state.get_index()

    def get_ground_atoms(self, skip_static = True, skip_fluent = False, skip_derived = False) -> 'list[GroundAtom]':
        """
        Returns the ground atoms of the state.
        """
        ground_atoms = []
        if not skip_static:
             static_ground_atoms = self._problem.get_initial_atoms(skip_static = False, skip_fluent = True, skip_derived = True)
             ground_atoms.extend([GroundAtom(x) for x in static_ground_atoms])
        if not skip_fluent:
            fluent_indices = self._advanced_state.get_fluent_atoms()
            fluent_ground_atoms = self._problem._advanced_problem.get_repositories().get_fluent_ground_atoms_from_indices(fluent_indices)
            ground_atoms.extend([GroundAtom(x) for x in fluent_ground_atoms])
        if not skip_derived:
            derived_indices = self._advanced_state.get_derived_atoms()
            derived_ground_atoms = self._problem._advanced_problem.get_repositories().get_derived_ground_atoms_from_indices(derived_indices)
            ground_atoms.extend([GroundAtom(x) for x in derived_ground_atoms])
        return ground_atoms

    def literal_holds(self, literal: GroundLiteral) -> bool:
        """
        Checks if a literal holds in the state.
        """
        return self._advanced_state.literal_holds(literal)

    def literals_hold(self, atoms: list[GroundLiteral]) -> bool:
        """
        Checks if all literals hold in the state.
        """
        return self._advanced_state.literals_hold(atoms)

    def __str__(self) -> 'str':
        """
        Returns a string representation of the state.
        """
        return self._advanced_state.to_string(self._problem._advanced_problem)

    def __repr__(self) -> 'str':
        """
        Returns a string representation of the state.
        """
        return self.__str__()

    def __hash__(self) -> 'int':
        """
        Returns a hash of the state.
        """
        return hash(self._advanced_state)

    def __eq__(self, other) -> bool:
        """
        Checks if two states are equal.
        """
        if not isinstance(other, State):
            return False
        return self._advanced_state == other._advanced_state


class ApplicableActionGenerator:
    """
    Class to generate applicable actions for a given state.
    """
    def __init__(self, problem: Problem):
        self._problem = problem

    def get_applicable_actions(self, state: State) -> 'list[Action]':
        """
        Returns the applicable actions for the given state.
        """
        return self._problem.get_applicable_actions(state._advanced_state)


class GroundConjunctiveCondition:
    _advanced_ground_conjunctive_condition: AdvancedGroundConjunctiveCondition = None

    def __init__(self, advanced_ground_conjunctive_condition: AdvancedGroundConjunctiveCondition) -> None:
        assert isinstance(advanced_ground_conjunctive_condition, AdvancedGroundConjunctiveCondition), "Invalid ground conjunctive condition type."
        self._advanced_ground_conjunctive_condition = advanced_ground_conjunctive_condition

    def holds(self, state: State) -> bool:
        """Check if the ground conjunctive condition holds in the given state."""
        assert isinstance(state, State), "Invalid state type."
        static_pos = self._advanced_ground_conjunctive_condition.get_static_positive_condition()
        static_neg = self._advanced_ground_conjunctive_condition.get_static_negative_condition()
        fluent_pos = self._advanced_ground_conjunctive_condition.get_fluent_positive_condition()
        fluent_neg = self._advanced_ground_conjunctive_condition.get_fluent_negative_condition()
        derived_pos = self._advanced_ground_conjunctive_condition.get_derived_positive_condition()
        derived_neg = self._advanced_ground_conjunctive_condition.get_derived_negative_condition()
        raise NotImplementedError("Not implemented.")
        # return state.literal_holds(static_pos) and \
        #        state.literal_holds(static_neg) and \
        #        state.literal_holds(fluent_pos) and \
        #        state.literal_holds(fluent_neg) and \
        #        state.literal_holds(derived_pos) and \
        #        state.literal_holds(derived_neg)


class ConjunctiveCondition:
    _advanced_conjunctive_condition: AdvancedConjunctiveCondition = None

    def __init__(self, advanced_condition: AdvancedConjunctiveCondition) -> None:
        assert isinstance(advanced_condition, AdvancedConjunctiveCondition), "Invalid conjunctive condition type."
        self._advanced_conjunctive_condition = advanced_condition

    def get_parameters(self) -> 'list[Variable]':
        """Get the parameters of the conjunctive condition."""
        return [Variable(x) for x in self._advanced_conjunctive_condition.get_parameters()]

    def get_literals(self, skip_static = False, skip_fluent = False, skip_derived = False) -> 'list[Literal]':
        """Get the literals of the conjunctive condition."""
        literals = []
        if not skip_static:
            literals.extend([Literal(x) for x in self._advanced_conjunctive_condition.get_static_literals()])
        if not skip_fluent:
            literals.extend([Literal(x) for x in self._advanced_conjunctive_condition.get_fluent_literals()])
        if not skip_derived:
            literals.extend([Literal(x) for x in self._advanced_conjunctive_condition.get_derived_literals()])
        return literals

    def get_nullary_ground_literals(self, skip_static = False, skip_fluent = False, skip_derived = False) -> 'list[GroundLiteral]':
        """Get the nullary ground literals of the conjunctive condition."""
        ground_literals = []
        if not skip_static:
            ground_literals.extend([GroundLiteral(x) for x in self._advanced_conjunctive_condition.get_nullary_ground_static_literals()])
        if not skip_fluent:
            ground_literals.extend([GroundLiteral(x) for x in self._advanced_conjunctive_condition.get_nullary_ground_fluent_literals()])
        if not skip_derived:
            ground_literals.extend([GroundLiteral(x) for x in self._advanced_conjunctive_condition.get_nullary_ground_derived_literals()])
        return ground_literals

    def __str__(self):
        """Get the string representation of the conjunctive condition."""
        return str(self._advanced_conjunctive_condition)

    def __repr__(self):
        """Get the string representation of the conjunctive condition."""
        return repr(self._advanced_conjunctive_condition)

    def __hash__(self) -> 'int':
        """Get the hash of the conjunctive condition."""
        return hash(self._advanced_conjunctive_condition)

    def __eq__(self, other: object) -> bool:
        """Check if two conjunctive conditions are equal."""
        if not isinstance(other, ConjunctiveCondition):
            return False
        return self._advanced_conjunctive_condition == other._advanced_conjunctive_condition

    # def get_numeric_constraints(self) -> 'list[NumericConstraint]':
    #     """Get the numeric constraints of the conjunctive condition."""
    #     return list(self._advanced_conjunctive_condition.get_numeric_constraints())



# Import all classes for better IDE support

# Enums
from pymimir.pymimir.advanced.search import (
    SearchMode,
    SearchNodeStatus,
    SearchStatus,
    MatchTreeSplitMetric,
    MatchTreeSplitStrategy,
    MatchTreeOptimizationDirection,
)

# Common
from pymimir.pymimir.advanced.search import (   
    BlindHeuristic,
    IApplicableActionGenerator,
    IAxiomEvaluator,
    IHeuristic,
    Plan,
    SearchResult,
    State,
    StateList,
    StateRepository,
    SearchContext,
    SearchContextOptions,
    GeneralizedSearchContext,
)

# SatisficingBindingGenerator
from pymimir.pymimir.advanced.search import (
    ActionSatisficingBindingGenerator,
    AxiomSatisficingBindingGenerator,
    ConjunctiveConditionSatisficingBindingGenerator
)

# GoalStrategy
from pymimir.pymimir.advanced.search import (
    IGoalStrategy,
    ProblemGoalStrategy,

    IPruningStrategy,
    NoPruningStrategy,
    DuplicatePruningStrategy,
    ArityZeroNoveltyPruningStrategy,
    ArityKNoveltyPruningStrategy,
)

# AStar
from pymimir.pymimir.advanced.search import (
    AStarStatistics,
    IAStarEventHandler,
    DebugAStarEventHandler,
    DefaultAStarEventHandler,
    find_solution_astar,
)

# BrFs
from pymimir.pymimir.advanced.search import (
    BrFSStatistics,
    IBrFSEventHandler,
    DebugBrFSEventHandler,
    DefaultBrFSEventHandler,
    find_solution_brfs,
)

# IW
from pymimir.pymimir.advanced.search import (
    IWStatistics,
    IIWEventHandler,
    DefaultIWEventHandler,
    find_solution_iw,

    TupleIndexMapper,
    StateTupleIndexGenerator,
    StatePairTupleIndexGenerator,
)

# SIW
from pymimir.pymimir.advanced.search import (
    SIWStatistics,
    ISIWEventHandler,
    DefaultSIWEventHandler,
    find_solution_siw,
)

# Lifted
from pymimir.pymimir.advanced.search import (
    DebugLiftedApplicableActionGeneratorEventHandler,
    DefaultLiftedApplicableActionGeneratorEventHandler,
    LiftedApplicableActionGenerator,
    LiftedAxiomEvaluator,
    ILiftedApplicableActionGeneratorEventHandler,
    ILiftedAxiomEvaluatorEventHandler,
)

# Grounded
from pymimir.pymimir.advanced.search import (
    DebugGroundedApplicableActionGeneratorEventHandler,
    DefaultGroundedApplicableActionGeneratorEventHandler,
    GroundedApplicableActionGenerator,
    GroundedAxiomEvaluator,
    IGroundedApplicableActionGeneratorEventHandler,
    IGroundedAxiomEvaluatorEventHandler,
    DeleteRelaxedProblemExplorator,
    MatchTreeOptions,
)




from pymimir.pymimir.advanced.languages.general_policies import (
    NamedConcept,
    NamedConceptList,
    NamedRole,
    NamedRoleList,
    NamedBoolean,
    NamedBooleanList,
    NamedNumerical,
    NamedNumericalList,

    Condition,
    ConditionList,
    PositiveBooleanCondition,
    NegativeBooleanCondition,
    GreaterNumericalCondition,
    EqualNumericalCondition,

    Effect,
    EffectList,
    PositiveBooleanEffect,
    NegativeBooleanEffect,
    UnchangedBooleanEffect,
    IncreaseNumericalEffect,
    DecreaseNumericalEffect,
    UnchangedNumericalEffect,

    Rule,
    RuleList,
    
    SolvabilityStatus,
    GeneralPolicy,

    Repositories,

    GeneralPolicyFactory,
)


import pymimir.advanced.formalism as formalism
import pymimir.advanced.search as search

from pathlib import Path

ROOT_DIR = (Path(__file__).parent.parent.parent.parent).absolute()


class CustomBlindHeuristic(search.IHeuristic):
    def __init__(self):
        super().__init__()

    def compute_heuristic(self, state : search.State, is_goal_state: bool) -> float:
        return 0.


class CustomEventHandler(search.IAStarEventHandler):
    def __init__(self):
        super().__init__()

        self.statistics = search.AStarStatistics()

    def on_expand_state(self, state : search.State):
        pass

    def on_expand_goal_state(self, state : search.State):
        pass

    def on_generate_state(self, state : search.State, action : formalism.GroundAction, action_cost: float, successor_state: search.State):
        pass

    def on_generate_state_relaxed(self, state : search.State, action : formalism.GroundAction, action_cost: float, successor_state: search.State):
        pass

    def on_generate_state_not_relaxed(self, state : search.State, action : formalism.GroundAction, action_cost: float, successor_state: search.State):
        pass

    def on_close_state(self, state : search.State):
        pass

    def on_finish_f_layer(self, f_value: float):
        pass

    def on_prune_state(self, state : search.State):
        pass

    def on_start_search(self, start_state : search.State):
        pass

    def on_end_search(self, num_reached_fluent_atoms : int, num_reached_derived_atoms: int, num_bytes_for_problem: int, num_bytes_for_nodes: int, num_states: int, num_nodes: int, num_actions: int, num_axioms: int):
        pass

    def on_solved(self, plan: search.Plan):
        pass

    def on_unsolvable(self):
        pass

    def on_exhausted(self):
        pass

    def get_statistics(self):
        return self.statistics

def test_astar_search():
    """ Test parsing a PDDL domain and problem file.
    """
    domain_filepath = str(ROOT_DIR / "data" / "gripper" / "domain.pddl")
    problem_filepath = str(ROOT_DIR / "data" / "gripper" / "test_problem.pddl")
    search_context = search.SearchContext.create(domain_filepath, problem_filepath)

    blind_heuristic = CustomBlindHeuristic()
    initial_state, _ = search_context.get_state_repository().get_or_create_initial_state()
    event_handler = CustomEventHandler()
    problem_goal_stategy = search.ProblemGoalStrategy(search_context.get_problem())
    no_pruning_strategy = search.NoPruningStrategy()

    result = search.find_solution_astar(search_context, blind_heuristic, initial_state, event_handler, problem_goal_stategy, no_pruning_strategy)

    assert result.status == search.SearchStatus.SOLVED
    assert len(result.plan) == 3



""" This example illustrates the interface of the central component for data generation.
"""

import pymimir.advanced.search as search
import pymimir.advanced.datasets as datasets

from pathlib import Path

ROOT_DIR = (Path(__file__).parent.parent.parent).absolute()


def main():
    domain_filepath = str(ROOT_DIR / "data" / "gripper" / "domain.pddl")
    problem_filepath_1 = str(ROOT_DIR / "data" / "gripper" / "p-1-0.pddl")
    problem_filepath_2 = str(ROOT_DIR / "data" / "gripper" / "p-2-0.pddl")

    # Create GeneralizedSearchContext
    search_context_options = search.SearchContextOptions()
    search_context_options.mode = search.SearchMode.GROUNDED
    generalized_search_context = search.GeneralizedSearchContext.create(domain_filepath, [problem_filepath_1, problem_filepath_2], search_context_options)

    # Create KnowledgeBase
    state_space_options = datasets.StateSpaceOptions()
    state_space_options.symmetry_pruning = False 
    generalized_state_space_options = datasets.GeneralizedStateSpaceOptions()
    tuple_graph_options = datasets.TupleGraphOptions()
    tuple_graph_options.width = 1
    knowledge_base_options = datasets.KnowledgeBaseOptions()
    knowledge_base_options.state_space_options = state_space_options
    knowledge_base_options.generalized_state_space_options = generalized_state_space_options
    knowledge_base_options.tuple_graph_options = tuple_graph_options
    knowledge_base = datasets.KnowledgeBase.create(generalized_search_context, knowledge_base_options)

    # 1. Access the StateSpaces.
    for state_space in knowledge_base.get_state_spaces():
        graph = state_space.get_graph()
        print("Dot representation of the StateSpace:")
        print(graph)

        # The graph is a specific instantiation of static graph which provides access to boost bgl. 
        # For more information on provided functionality, see:
        # https://github.com/simon-stahlberg/mimir/blob/main/python/examples/static_graph.py
        # Since the graph is a specific instantiation, the vertices and edges contain context-specific information.
        # Iterate over vertices
        for vertex in graph.get_vertices():
            # Access vertex properties through get_property_i or specific free function
            vertex.get_index()      
            datasets.get_state(vertex)
            datasets.get_problem(vertex)
            datasets.get_unit_goal_distance(vertex)
            datasets.get_action_goal_distance(vertex)
            datasets.is_initial(vertex)
            datasets.is_goal(vertex)
            datasets.is_unsolvable(vertex)
            datasets.is_alive(vertex)
        # Iterate over edges
        for edge in graph.get_edges():
            # Access edge properties through get_property_i or specific free function
            edge.get_index()  
            edge.get_source()   
            edge.get_target()   
            datasets.get_action(edge)
            datasets.get_problem(edge)
            datasets.get_action_cost(edge)
    

    # 2. Access GeneralizedStateSpace
    generalized_state_space = knowledge_base.get_generalized_state_space()

    # Access GeneralizedStateSpace meta data
    generalized_state_space.get_initial_vertices()
    generalized_state_space.get_unsolvable_vertices()
    generalized_state_space.get_goal_vertices()

    # Access GeneralizedStateSpace graph
    graph = generalized_state_space.get_graph()
    print("Dot represention of the GeneralizedStateSpace:")
    print(graph)

    # The graph is a specific instantiation of static graph which provides access to boost bgl. 
    # For more information on provided functionality, see:
    # https://github.com/simon-stahlberg/mimir/blob/main/python/examples/static_graph.py
    # Since the graph is a specific instantiation, the vertices and edges contain context-specific information.
    # Iterate over vertices
    for vertex in graph.get_vertices():
        # Access vertex properties get_property_i or specific free function
        vertex.get_index()      
        datasets.get_problem_vertex_index(vertex)
        datasets.get_problem_index(vertex)
    # Iterate over edges
    for edge in graph.get_edges():
        # Access edge properties get_property_i or specific free function
        edge.get_index()  
        edge.get_source()   
        edge.get_target()   
        datasets.get_problem_edge_index(edge)
        datasets.get_problem_index(edge)

    # The graph also represents an abstraction of a collection of state spaces.
    # Access low-level information about state and actions using the mappings encoded in the GeneralizedStateSpace.
    for vertex in graph.get_vertices():
        # Map class vertex to problem vertex
        problem_vertex = generalized_state_space.get_problem_vertex(vertex)
        # Acces vertex properties get_property_i or specific free function
        problem_vertex.get_index()
        datasets.get_state(problem_vertex)
        datasets.get_problem(problem_vertex)
        datasets.get_unit_goal_distance(problem_vertex)
        datasets.get_action_goal_distance(problem_vertex)
        datasets.is_initial(problem_vertex)
        datasets.is_goal(problem_vertex)
        datasets.is_unsolvable(problem_vertex)
        datasets.is_alive(problem_vertex)
    for edge in graph.get_edges():
        # Map class edge to problem edge
        problem_edge = generalized_state_space.get_problem_edge(edge)
        # Access edge properties get_property_i or specific free function
        problem_edge.get_index()
        problem_edge.get_source()
        problem_edge.get_target()
        datasets.get_action(problem_edge)
        datasets.get_problem(problem_edge)
        datasets.get_action_cost(problem_edge)

    # 3. Access the tuple graphs of each state space.
    for tuple_graphs in knowledge_base.get_tuple_graphs():
        # Access the tuple graph in the state space.
        for tuple_graph in tuple_graphs:
            tuple_graph.get_graph()
            tuple_graph.get_state_space()
            tuple_graph.get_problem_vertex_indices_grouped_by_distance()
            tuple_graph.get_tuple_vertex_indices_grouped_by_distance()

            for vertex in tuple_graph.get_graph().get_vertices():
                vertex.get_index()
                datasets.get_atom_tuple(vertex)
                datasets.get_problem_vertices(vertex)


if __name__ == "__main__":
    main()




# Enums
from pymimir.pymimir.advanced.languages.description_logics import (
    GrammarSpecificationEnum,
)

# DL constructors
from pymimir.pymimir.advanced.languages.description_logics import (
    Concept,
    ConceptList,
    Role,
    RoleList,
    Boolean,
    BooleanList,
    Numerical,
    NumericalList,
    ConceptBot,
    ConceptTop,
    ConceptStaticAtomicState,
    ConceptFluentAtomicState,
    ConceptDerivedAtomicState,
    ConceptStaticAtomicGoal,
    ConceptFluentAtomicGoal,
    ConceptDerivedAtomicGoal,
    ConceptIntersection,
    ConceptUnion,
    ConceptNegation,
    ConceptValueRestriction,
    ConceptExistentialQuantification,
    ConceptRoleValueMapContainment,
    ConceptRoleValueMapEquality,
    ConceptNominal,
    RoleUniversal,
    RoleStaticAtomicState,
    RoleFluentAtomicState,
    RoleDerivedAtomicState,
    RoleStaticAtomicGoal,
    RoleFluentAtomicGoal,
    RoleDerivedAtomicGoal,
    RoleIntersection,
    RoleUnion,
    RoleComplement,
    RoleInverse,
    RoleComposition,
    RoleTransitiveClosure,
    RoleReflexiveTransitiveClosure,
    RoleRestriction,
    RoleIdentity,
    BooleanStaticAtomicState,
    BooleanFluentAtomicState,
    BooleanDerivedAtomicState,
    BooleanConceptNonempty,
    BooleanRoleNonempty,
    NumericalConceptCount,
    NumericalRoleCount,
    NumericalDistance,

    Repositories,

    ConceptDenotation,
    RoleDenotation,
    BooleanDenotation,
    NumericalDenotation,
    DenotationRepositories,
    EvaluationContext,
)

# CNF grammar

from pymimir.pymimir.advanced.languages.description_logics import (
    CNFConcept,
    CNFRole,
    CNFBoolean,
    CNFNumerical,
    CNFConceptNonTerminal,
    CNFRoleNonTerminal,
    CNFBooleanNonTerminal,
    CNFNumericalNonTerminal,
    CNFConceptDerivationRule,
    CNFRoleDerivationRule,
    CNFBooleanDerivationRule,
    CNFNumericalDerivationRule,
    CNFConceptSubstitutionRule,
    CNFRoleSubstitutionRule,
    CNFBooleanSubstitutionRule,
    CNFNumericalSubstitutionRule,
    CNFConceptBot,
    CNFConceptTop,
    CNFConceptStaticAtomicState,
    CNFConceptFluentAtomicState,
    CNFConceptDerivedAtomicState,
    CNFConceptStaticAtomicGoal,
    CNFConceptFluentAtomicGoal,
    CNFConceptDerivedAtomicGoal,
    CNFConceptIntersection,
    CNFConceptUnion,
    CNFConceptNegation,
    CNFConceptValueRestriction,
    CNFConceptExistentialQuantification,
    CNFConceptRoleValueMapContainment,
    CNFConceptRoleValueMapEquality,
    CNFConceptNominal,
    CNFRoleUniversal,
    CNFRoleStaticAtomicState,
    CNFRoleFluentAtomicState,
    CNFRoleDerivedAtomicState,
    CNFRoleStaticAtomicGoal,
    CNFRoleFluentAtomicGoal,
    CNFRoleDerivedAtomicGoal,
    CNFRoleIntersection,
    CNFRoleUnion,
    CNFRoleComplement,
    CNFRoleInverse,
    CNFRoleComposition,
    CNFRoleTransitiveClosure,
    CNFRoleReflexiveTransitiveClosure,
    CNFRoleRestriction,
    CNFRoleIdentity,
    CNFBooleanStaticAtomicState,
    CNFBooleanFluentAtomicState,
    CNFBooleanDerivedAtomicState,
    CNFBooleanConceptNonempty,
    CNFBooleanRoleNonempty,
    CNFNumericalConceptCount,
    CNFNumericalRoleCount,
    CNFNumericalDistance,

    CNFRepositories,

    CNFGrammar,
    StateListRefinementPruningFunction,
    GeneratedSentencesContainer,
    GeneratorVisitor,
)


""" This example illustrates how to create traces string representations of actions and objects 
    that may be obtained from plans of external planners.
"""

import pymimir.advanced.search as search
import pymimir.advanced.formalism as formalism

from pathlib import Path
from typing import Dict

ROOT_DIR = (Path(__file__).parent.parent.parent).absolute()


def main():
    domain_filepath = str(ROOT_DIR / "data" / "gripper" / "domain.pddl")
    problem_filepath = str(ROOT_DIR / "data" / "gripper" / "test_problem.pddl")
    problem = formalism.Problem.create(domain_filepath, problem_filepath)

    actions_by_name: Dict[str, formalism.GroundAction] = dict()
    for action in problem.get_domain().get_actions():
        actions_by_name[action.get_name()] = action
    objects_by_name: Dict[str, formalism.Object] = dict()
    for object in problem.get_problem_and_domain_objects():
        objects_by_name[object.get_name()] = object

    applicable_action_generator = search.LiftedApplicableActionGenerator(problem)
    axiom_evaluator = search.LiftedAxiomEvaluator(problem)
    state_repository = search.StateRepository(axiom_evaluator)
    problem_goal_strategy = search.ProblemGoalStrategy(problem)

    # Empty static goal is trivially satisfied
    assert(problem_goal_strategy.test_static_goal())

    initial_state, initial_metric_value = state_repository.get_or_create_initial_state()
    print(initial_state.to_string(problem))
    assert(not problem_goal_strategy.test_dynamic_goal(initial_state))

    move_rooma_roomb_action = problem.ground(actions_by_name["move"], formalism.ObjectList([objects_by_name["rooma"], objects_by_name["roomb"]]))
    print(move_rooma_roomb_action.to_string(problem))

    successor_state, successor_metric_value = state_repository.get_or_create_successor_state(initial_state, move_rooma_roomb_action, initial_metric_value)
    print(successor_state.to_string(problem))
    assert(not problem_goal_strategy.test_dynamic_goal(successor_state))


if __name__ == "__main__":
    main()



""" This example illustrates the interface of a generic static graph taking arbitrary python tuples as vertex and edge property.
    Static graphs dont allow vertex or edge removal allowing for most efficient data representation. 
"""

import pymimir.advanced.graphs as graphs


def main():
    graph = graphs.StaticPyGraph()
    v0 = graph.add_vertex((0,0))
    v1 = graph.add_vertex((0,1))
    v2 = graph.add_vertex((1,1))
    v3 = graph.add_vertex((1,0))
    graph.add_undirected_edge(v0, v1, ("a",1))
    graph.add_undirected_edge(v1, v2, ("b",2))
    graph.add_undirected_edge(v2, v3, ("c",3))
    graph.add_undirected_edge(v3, v0, ("d",4))

    # Print dot representation of the graph.

    print(graph)

    # Check properties of the graph.

    assert(graphs.is_undirected(graph))
    assert(graphs.is_loopless(graph))
    assert(not graphs.is_acyclic(graph))
    assert(not graphs.is_multi(graph))

    # Translate to immutable forward and bidirectional traversal optimized graphs 
    # relevant for algorithms that traverse vertices or edges adjacent to a given vertex.
    forward_graph = graphs.ForwardStaticPyGraph(graph)
    bidirectional_graph = graphs.BidirectionalStaticPyGraph(graph)

    # Run basic graph algorithms with switchable edge direction

    edge_weights = [1] * graph.get_num_edges()
    source_indices = [0,1]

    graph.compute_strong_components()

    graph.compute_forward_breadth_first_search(source_indices)
    graph.compute_forward_depth_first_search(source_indices)
    graph.compute_forward_floyd_warshall_all_pairs_shortest_paths(edge_weights)
    graph.compute_forward_dijkstra_shortest_paths(edge_weights, source_indices)
    try:
        graph.compute_forward_topological_sort()  # graph is cyclic => this will throw ValueError
    except ValueError:
        pass
    
    graph.compute_backward_breadth_first_search(source_indices)
    graph.compute_backward_depth_first_search(source_indices)
    graph.compute_backward_floyd_warshall_all_pairs_shortest_paths(edge_weights)
    graph.compute_backward_dijkstra_shortest_paths(edge_weights, source_indices)
    try:
        graph.compute_backward_topological_sort()  # graph is cyclic => this will throw ValueError
    except ValueError:
        pass

    # Iterate over vertices
    for vertex in graph.get_vertices():
        # Access vertex properties through generic member
        vertex.get_index()      
        vertex.get_property_0()  # python tuple

        # Iterate over adjacent structures
        for adj_vertex in graph.get_forward_adjacent_vertices(vertex.get_index()):
            pass

        for adj_vertex in graph.get_backward_adjacent_vertices(vertex.get_index()):
            pass 

        for adj_edge in graph.get_forward_adjacent_edges(vertex.get_index()):
            pass 

        for adj_edge in graph.get_backward_adjacent_edges(vertex.get_index()):
            pass 

        for adj_v_idx in graph.get_forward_adjacent_vertex_indices(vertex.get_index()):
            pass 

        for adj_v_idx in graph.get_backward_adjacent_vertex_indices(vertex.get_index()):
            pass 

        for adj_e_idx in graph.get_forward_adjacent_edge_indices(vertex.get_index()):
            pass 

        for adj_e_idx in graph.get_backward_adjacent_edge_indices(vertex.get_index()):
            pass
            
    # Iterate over edges
    for edge in graph.get_edges():
        # Access edge properties through generic member
        edge.get_index()  
        edge.get_source()   
        edge.get_target()   
        edge.get_property_0()  # python tuple



if __name__ == "__main__":
    main()


""" This example illustrates the interface for graph isomorphism testing of undirected vertex colored graphs 
    using a wrapper around nauty and Traces (https://pallini.di.uniroma1.it/).
"""

import pymimir.advanced.graphs as graphs 

def main():
    graph1 = graphs.StaticVertexColoredGraph()
    v1_1 = graph1.add_vertex(graphs.Color(1, "a"))
    v2_1 = graph1.add_vertex(graphs.Color(2, "b"))
    v3_1 = graph1.add_vertex(graphs.Color(3, "c"))
    graph1.add_undirected_edge(v1_1, v3_1)
    print(graph1)  # print dot representation of graph1
 
    # graph2 is isomorphic to graph1 but it is vertices are added in a different order.
    graph2 = graphs.StaticVertexColoredGraph()
    v1_2 = graph2.add_vertex(graphs.Color(2, "b"))
    v2_2 = graph2.add_vertex(graphs.Color(1, "a"))
    v3_2 = graph2.add_vertex(graphs.Color(3, "c"))
    graph2.add_undirected_edge(v2_2, v3_2)
    print(graph2)  # print dot representation of graph2

    # Exact graph isomorphism via nauty with Traces
    nauty_graph1 = graphs.NautySparseGraph(graph1)
    nauty_graph2 = graphs.NautySparseGraph(graph2)
    assert(nauty_graph1 != nauty_graph2)
    nauty_graph1.canonize()
    nauty_graph2.canonize()  
    assert(nauty_graph1 == nauty_graph2)
 
    # Approximate graph isomorphism via k-dimensional Weisfeiler-Leman (k-WL) algorithm
    stable_coloring_1 = graphs.compute_color_refinement_certificate(graph1)
    stable_coloring_2 = graphs.compute_color_refinement_certificate(graph2)
    assert(stable_coloring_1 == stable_coloring_2)
    isomorphism_type_compression_function = graphs.KFWLIsomorphismTypeCompressionFunction()
    stable_coloring_1 = graphs.compute_2fwl_certificate(graph1, isomorphism_type_compression_function)
    stable_coloring_2 = graphs.compute_2fwl_certificate(graph2, isomorphism_type_compression_function)
    assert(stable_coloring_1 == stable_coloring_2)
    stable_coloring_1 = graphs.compute_3fwl_certificate(graph1, isomorphism_type_compression_function)
    stable_coloring_2 = graphs.compute_3fwl_certificate(graph2, isomorphism_type_compression_function)
    assert(stable_coloring_1 == stable_coloring_2)
    stable_coloring_1 = graphs.compute_4fwl_certificate(graph1, isomorphism_type_compression_function)
    stable_coloring_2 = graphs.compute_4fwl_certificate(graph2, isomorphism_type_compression_function)
    assert(stable_coloring_1 == stable_coloring_2)

if __name__ == "__main__":
    main()



# Import all classes for better IDE support

# StateSpace
from pymimir.pymimir.advanced.datasets import (
    ProblemVertex,
    get_state,
    get_problem,
    get_unit_goal_distance,
    get_action_goal_distance,
    is_initial,
    is_goal,
    is_unsolvable,
    is_alive,
    ProblemEdge,
    get_action,
    get_problem,
    get_action_cost,
    StaticProblemGraph,
    ImmutableStaticProblemGraph,
    ForwardStaticProblemGraph,
    BidirectionalStaticProblemGraph,

    StateSpace,
    StateSpaceOptions,
)

# GeneralizedStateSpace
from pymimir.pymimir.advanced.datasets import (
    ClassVertex,
    get_problem_vertex_index,
    get_problem_index,
    ClassEdge,
    get_problem_edge_index,
    get_problem_index,
    StaticClassGraph,
    ImmutableStaticClassGraph,
    ForwardStaticClassGraph,
    BidirectionalStaticClassGraph,

    GeneralizedStateSpace,
    GeneralizedStateSpaceOptions,
)

# TupleGraph
from pymimir.pymimir.advanced.datasets import (
    TupleGraphVertex,
    get_atom_tuple,
    get_problem_vertices,
    StaticTupleGraph,
    ImmutableStaticTupleGraph,
    ForwardStaticTupleGraph,
    BidirectionalStaticTupleGraph,

    TupleGraphOptions,
    TupleGraph,
)

# KnowledgeBase
from pymimir.pymimir.advanced.datasets import (
    KnowledgeBase,
    KnowledgeBaseOptions,
)

# ObjectGraph 
from pymimir.pymimir.advanced.datasets import (
    create_object_graph
)

__all__ = [
    # StateSpace 
    ProblemVertex,
    ProblemEdge,
    StaticProblemGraph,
    ImmutableStaticProblemGraph,
    ForwardStaticProblemGraph,
    BidirectionalStaticProblemGraph,

    StateSpace,
    StateSpaceOptions,

    # GeneralizedStateSpace
    ClassVertex,
    ClassEdge,
    StaticClassGraph,
    ImmutableStaticClassGraph,
    ForwardStaticClassGraph,
    BidirectionalStaticClassGraph,

    GeneralizedStateSpace,
    GeneralizedStateSpaceOptions,

    # TupleGraph
    TupleGraphVertex,
    StaticTupleGraph,
    ImmutableStaticTupleGraph,
    ForwardStaticTupleGraph,
    BidirectionalStaticTupleGraph,

    TupleGraphOptions,
    TupleGraph,

    # KnowledgeBase
    KnowledgeBase,
    KnowledgeBaseOptions,

    # ObjectGraph
    create_object_graph
]


# Import all classes for better IDE support

from pymimir.pymimir.advanced.common import (
    FlatBitset,
    FlatDoubleList,
    FlatIndexList,
)

__all__ = ["FlatBitset", "FlatDoubleList", "FlatIndexList"]



import pymimir.advanced.formalism as formalism

from pathlib import Path

ROOT_DIR = (Path(__file__).parent.parent.parent.parent).absolute()


def test_pddl_parser():
    """ Test parsing and translation of a PDDL domain and problem file.
    """
    domain_filepath = str(ROOT_DIR / "data" / "gripper" / "domain.pddl")
    problem_filepath = str(ROOT_DIR / "data" / "gripper" / "test_problem.pddl")
    parser = formalism.Parser(domain_filepath)

    domain = parser.get_domain()

    problem = parser.parse_problem(problem_filepath)

    assert(domain == problem.get_domain())

    translator = formalism.Translator(domain)
    translated_domain = translator.get_translated_domain()

    translated_problem = translator.translate(problem)

    assert(translated_domain == translated_problem.get_domain())



import unittest

from pathlib import Path
from wrapper import *


DATA_DIR = (Path(__file__).parent.parent.parent.parent).absolute() / 'data'


class TestDomain(unittest.TestCase):
    def test_name(self):
        domain_path = DATA_DIR / 'blocks_4' / 'domain.pddl'
        domain = Domain(domain_path)
        actual_name = domain.get_name()
        expected_name = 'blocksworld'
        assert actual_name == expected_name

    def test_requirements(self):
        domain_path = DATA_DIR / 'blocks_4' / 'domain.pddl'
        domain = Domain(domain_path)
        actual_requirements = domain.get_requirements()
        expected_requirements = [':strips']
        assert len(actual_requirements) == len(expected_requirements)
        for requirement in actual_requirements:
            assert requirement in expected_requirements

    def test_predicates(self):
        domain_path = DATA_DIR / 'blocks_4' / 'domain.pddl'
        domain = Domain(domain_path)
        actual_predicates = domain.get_predicates()
        expected_predicates = [('number', 1), ('object', 1), ('on', 2), ('clear', 1), ('holding', 1), ('on-table', 1), ('arm-empty', 0)]
        assert len(actual_predicates) == len(expected_predicates)
        for predicate in actual_predicates:
            assert predicate.get_index() is not None
            assert (predicate.get_name(), predicate.get_arity()) in expected_predicates

    def test_actions(self):
        domain_path = DATA_DIR / 'blocks_4' / 'domain.pddl'
        domain = Domain(domain_path)
        actual_actions = domain.get_actions()
        expected_actions = [('pickup', 1), ('putdown', 1), ('stack', 2), ('unstack', 2)]
        assert len(actual_actions) == len(expected_actions)
        for action in actual_actions:
            assert action.get_index() is not None
            assert (action.get_name(), action.get_arity()) in expected_actions

    def test_constants(self):
        domain_path = DATA_DIR / 'woodworking' / 'domain.pddl'
        domain = Domain(domain_path)
        actual_constants = domain.get_constants()
        expected_constants = ['verysmooth', 'smooth', 'rough', 'varnished', 'glazed', 'untreated', 'colourfragments', 'natural', 'small', 'medium', 'large']
        assert len(actual_constants) == len(expected_constants)
        for constant in actual_constants:
            assert constant.get_index() is not None
            assert constant.get_name() in expected_constants


class TestProblem(unittest.TestCase):
    def test_name(self):
        domain_path = DATA_DIR / 'blocks_4' / 'domain.pddl'
        problem_path = DATA_DIR / 'blocks_4' / 'test_problem.pddl'
        domain = Domain(domain_path)
        problem = Problem(domain, problem_path)
        actual_name = problem.get_name()
        expected_name = 'blocksworld-300'
        assert problem.get_index() is not None
        assert actual_name == expected_name

    # def test_requirements(self):
    #     domain_path = DATA_DIR / 'blocks_4' / 'domain.pddl'
    #     problem_path = DATA_DIR / 'blocks_4' / 'test_problem.pddl'
    #     domain = Domain(domain_path)
    #     problem = Problem(domain, problem_path)
    #     actual_requirements = problem.get_requirements()
    #     expected_requirements = [':strips', ':typing'] # TODO: Dominik -- :typing is missing in the requirements of the problem.
    #     assert len(actual_requirements) == len(expected_requirements)
    #     for requirement in actual_requirements:
    #         assert requirement in expected_requirements

    def test_objects(self):
        domain_path = DATA_DIR / 'blocks_4' / 'domain.pddl'
        problem_path = DATA_DIR / 'blocks_4' / 'test_problem.pddl'
        domain = Domain(domain_path)
        problem = Problem(domain, problem_path)
        actual_objects = problem.get_objects()
        expected_objects = ['b1', 'b2', 'b3']
        assert len(actual_objects) == len(expected_objects)
        for obj in actual_objects:
            assert obj.get_index() is not None
            assert obj.get_name() in expected_objects

    def test_initial_atoms(self):
        domain_path = DATA_DIR / 'blocks_4' / 'domain.pddl'
        problem_path = DATA_DIR / 'blocks_4' / 'test_problem.pddl'
        domain = Domain(domain_path)
        problem = Problem(domain, problem_path)
        actual_initial_atoms = problem.get_initial_atoms(skip_derived=True)  # TODO: Dominik -- Crashes with skip_derived=False
        expected_initial_atoms = ['(arm-empty)', '(clear b2)', '(on-table b2)', '(clear b1)', '(on b1 b3)', '(on-table b3)']
        assert len(actual_initial_atoms) == len(expected_initial_atoms)
        for initial_atom in actual_initial_atoms:
            assert initial_atom.get_index() is not None
            assert str(initial_atom) in expected_initial_atoms

    def test_new_atom_and_literal(self):
        domain_path = DATA_DIR / 'blocks_4' / 'domain.pddl'
        problem_path = DATA_DIR / 'blocks_4' / 'test_problem.pddl'
        domain = Domain(domain_path)
        problem = Problem(domain, problem_path)
        predicate = domain.get_predicate('on')
        variables = problem.new_variable_list(['?v1', '?v2'])
        actual_new_atom = problem.new_atom(predicate, [variables[0], variables[1]])
        expected_new_atom = '(on ?v1 ?v2)'
        assert actual_new_atom.get_index() is not None
        assert str(actual_new_atom) == expected_new_atom
        actual_new_literal = problem.new_literal(actual_new_atom, False)
        expected_new_literal = '(not (on ?v1 ?v2))'
        assert actual_new_literal.get_index() is not None
        assert str(actual_new_literal) == expected_new_literal

    def test_new_ground_atom_and_ground_literal(self):
        domain_path = DATA_DIR / 'blocks_4' / 'domain.pddl'
        problem_path = DATA_DIR / 'blocks_4' / 'test_problem.pddl'
        domain = Domain(domain_path)
        problem = Problem(domain, problem_path)
        predicate = domain.get_predicate('on')
        objects = [problem.get_object('b1'), problem.get_object('b2')]
        actual_new_ground_atom = problem.new_ground_atom(predicate, [objects[0], objects[1]])
        expected_new_ground_atom = '(on b1 b2)'
        assert actual_new_ground_atom.get_index() is not None
        assert str(actual_new_ground_atom) == expected_new_ground_atom
        actual_new_ground_literal = problem.new_ground_literal(actual_new_ground_atom, False)
        expected_new_ground_literal = '(not (on b1 b2))'
        assert actual_new_ground_literal.get_index() is not None
        assert str(actual_new_ground_literal) == expected_new_ground_literal

    # def test_goal_condition(self):
    #     domain_path = DATA_DIR / 'blocks_4' / 'domain.pddl'
    #     problem_path = DATA_DIR / 'blocks_4' / 'test_problem.pddl'
    #     domain = Domain(domain_path)
    #     problem = Problem(domain, problem_path)
    #     actual_goal_condition = problem.get_goal_condition()  # TODO: Dominik -- Not implemented.
    #     expected_goal_condition = ['(clear b2)', '(on b2 b3)', '(on-table b3)', '(clear b1)', '(on-table b1)']
    #     assert len(actual_goal_condition) == len(expected_goal_condition)
    #     for goal_literal in actual_goal_condition:
    #         assert goal_literal.get_polarity() is True
    #         assert str(goal_literal.get_atom()) in expected_goal_condition


class TestAction(unittest.TestCase):
    def test_precondition(self):
        domain_path = DATA_DIR / 'miconic-fulladl' / 'domain.pddl'
        domain = Domain(domain_path)
        action = domain.get_action('up')
        assert action.get_name() == 'up'
        actual_precondition = action.get_conjunctive_condition()
        actual_parameters = actual_precondition.get_parameters()
        expected_parameters = ['?f1_0', '?f2_0']
        assert len(actual_parameters) == len(expected_parameters)
        for parameter in actual_parameters:
            assert parameter.get_index() is not None
            assert parameter.get_name() in expected_parameters
        actual_all_literals = actual_precondition.get_literals()
        expected_all_literals = ['(floor ?f1_0)', '(floor ?f2_0)', '(above ?f1_0 ?f2_0)', '(lift-at ?f1_0)', '(not (axiom_8))']
        assert len(actual_all_literals) == len(expected_all_literals)
        for literal in actual_all_literals:
            assert literal.get_index() is not None
            assert str(literal) in expected_all_literals
        actual_static_literals = actual_precondition.get_literals(False, True, True)
        expected_static_literals  = ['(floor ?f1_0)', '(floor ?f2_0)', '(above ?f1_0 ?f2_0)']
        assert len(actual_static_literals) == len(expected_static_literals)
        for literal in actual_static_literals:
            assert literal.get_index() is not None
            assert str(literal) in expected_static_literals
        actual_fluent_literals = actual_precondition.get_literals(True, False, True)
        expected_fluent_literals  = ['(lift-at ?f1_0)']
        assert len(actual_fluent_literals) == len(expected_fluent_literals)
        for literal in actual_fluent_literals:
            assert literal.get_index() is not None
            assert str(literal) in expected_fluent_literals
        actual_derived_literals = actual_precondition.get_literals(True, True, False)
        expected_derived_literals  = ['(not (axiom_8))']
        assert len(actual_derived_literals) == len(expected_derived_literals)
        for literal in actual_derived_literals:
            assert literal.get_index() is not None
            assert str(literal) in expected_derived_literals

    def test_effect(self):
        domain_path = DATA_DIR / 'miconic-fulladl' / 'domain.pddl'
        domain = Domain(domain_path)
        action = domain.get_action('up')
        assert action.get_name() == 'up'
        actual_effect = action.get_conjunctive_effect()
        actual_parameters = actual_effect.get_parameters()
        expected_parameters = ['?f1_0', '?f2_0']
        assert len(actual_parameters) == len(expected_parameters)
        for parameter in actual_parameters:
            assert parameter.get_index() is not None
            assert parameter.get_name() in expected_parameters
        actual_literals = actual_effect.get_literals()
        expected_literals = ['(lift-at ?f2_0)', '(not (lift-at ?f1_0))']
        assert len(actual_literals) == len(expected_literals)
        for literal in actual_literals:
            assert literal.get_index() is not None
            assert str(literal) in expected_literals

    def test_conditional_effect(self):
        domain_path = DATA_DIR / 'miconic-fulladl' / 'domain.pddl'
        domain = Domain(domain_path)
        action = domain.get_action('stop')
        assert action.get_name() == 'stop'
        actual_conditiona_effect = action.get_conditional_effect()
        pass



if __name__ == '__main__':
    unittest.main()



""" This example illustrates the interface of the description logics component.
"""

import pymimir.advanced.search as search
import pymimir.advanced.datasets as datasets
import pymimir.advanced.languages.description_logics as description_logics
import pymimir.advanced.languages.general_policies as general_policies

from functools import singledispatchmethod
from pathlib import Path

from description_logics import MyConstructorVisitor

ROOT_DIR = (Path(__file__).parent.parent.parent).absolute()


class MyGeneralPolicyVisitor:
    """ Recursively visit the general policy. """
    def __init__(self):
        self.constructor_visitor = MyConstructorVisitor()

    @singledispatchmethod
    def visit(self, arg):
        raise NotImplementedError("Missing implementation for the constructor!" )
    
    @visit.register
    def _(self, feature : general_policies.NamedConcept):
        self.constructor_visitor.visit(feature.get_feature())
    @visit.register
    def _(self, feature : general_policies.NamedRole):
        self.constructor_visitor.visit(feature.get_feature())
    @visit.register
    def _(self, feature : general_policies.NamedBoolean):
        self.constructor_visitor.visit(feature.get_feature())
    @visit.register
    def _(self, feature : general_policies.NamedNumerical):
        self.constructor_visitor.visit(feature.get_feature())
    @visit.register
    def _(self, condition : general_policies.PositiveBooleanCondition):
        self.visit(condition.get_feature())
    @visit.register
    def _(self, condition : general_policies.NegativeBooleanCondition):
        self.visit(condition.get_feature())
    @visit.register
    def _(self, condition : general_policies.GreaterNumericalCondition):
        self.visit(condition.get_feature())
    @visit.register
    def _(self, condition : general_policies.EqualNumericalCondition):
        self.visit(condition.get_feature())
    @visit.register
    def _(self, effect : general_policies.PositiveBooleanEffect):
        self.visit(effect.get_feature())
    @visit.register
    def _(self, effect : general_policies.NegativeBooleanEffect):
        self.visit(effect.get_feature())
    @visit.register
    def _(self, effect : general_policies.UnchangedBooleanEffect):
        self.visit(effect.get_feature())
    @visit.register
    def _(self, effect : general_policies.IncreaseNumericalEffect):
        self.visit(effect.get_feature())
    @visit.register
    def _(self, effect : general_policies.DecreaseNumericalEffect):
        self.visit(effect.get_feature())
    @visit.register
    def _(self, effect : general_policies.UnchangedNumericalEffect):
        self.visit(effect.get_feature())
    @visit.register
    def _(self, rule : general_policies.Rule):
        for condition in rule.get_conditions():
            self.visit(condition)
        for effect in rule.get_effects():
            self.visit(effect)
    @visit.register
    def _(self, policy : general_policies.GeneralPolicy):
        for boolean in policy.get_boolean_features():
            self.visit(boolean)
        for numerical in policy.get_boolean_features():
            self.visit(numerical)
        for rule in policy.get_rules():
            self.visit(rule)


def main():
    domain_filepath = str(ROOT_DIR / "data" / "gripper" / "domain.pddl")
    problem_filepath = str(ROOT_DIR / "data" / "gripper" / "p-1-0.pddl")

    search_context_options = search.SearchContextOptions()
    search_context_options.mode = search.SearchMode.GROUNDED
    generalized_search_context = search.GeneralizedSearchContext.create(domain_filepath, [problem_filepath], search_context_options)

    state_space_options = datasets.StateSpaceOptions()
    state_space_options.symmetry_pruning = True 
    generalized_state_space_options = datasets.GeneralizedStateSpaceOptions()
    knowledge_base_options = datasets.KnowledgeBaseOptions()
    knowledge_base_options.state_space_options = state_space_options
    knowledge_base_options.generalized_state_space_options = generalized_state_space_options
    knowledge_base = datasets.KnowledgeBase.create(generalized_search_context, knowledge_base_options)

    repositories = general_policies.Repositories()
    dl_repositories = description_logics.Repositories()
    denotation_repositories = description_logics.DenotationRepositories()

    # Create a policy from description.
    general_policy = repositories.get_or_create_general_policy(
        """
        [boolean_features]
            <boolean_r_b> ::= 
                @boolean_nonempty 
                    @concept_existential_quantification 
                        @role_atomic_goal "at" true 
                        @concept_atomic_state "at-robby"

        [numerical_features]
            <numerical_c> ::= 
                @numerical_count 
                    @concept_existential_quantification 
                        @role_atomic_state "carry" 
                        @concept_top
            <numerical_b> ::=
                @numerical_count 
                    @concept_negation
                        @concept_role_value_map_equality
                            @role_atomic_state "at"
                            @role_atomic_goal "at" true

        [policy_rules]
            { @negative_boolean_condition <boolean_r_b>, @equal_numerical_condition <numerical_c>, @greater_numerical_condition <numerical_b> } 
            -> { @unchanged_boolean_effect <boolean_r_b>, @unchanged_numerical_effect <numerical_b>, @increase_numerical_effect <numerical_c> }
            { @positive_boolean_condition <boolean_r_b>, @equal_numerical_condition <numerical_c>, @greater_numerical_condition <numerical_b> } 
            -> { @negative_boolean_effect <boolean_r_b>, @unchanged_numerical_effect <numerical_b>, @unchanged_numerical_effect <numerical_c> }
            { @positive_boolean_condition <boolean_r_b>, @greater_numerical_condition <numerical_c>, @greater_numerical_condition <numerical_b> } 
            -> { @unchanged_boolean_effect <boolean_r_b>, @decrease_numerical_effect <numerical_c>, @decrease_numerical_effect <numerical_b> }
            { @negative_boolean_condition <boolean_r_b>, @greater_numerical_condition <numerical_c>, @greater_numerical_condition <numerical_b> } 
            -> { @positive_boolean_effect <boolean_r_b>, @unchanged_numerical_effect <numerical_b>, @unchanged_numerical_effect <numerical_c> }
        """,
        knowledge_base.get_domain(), dl_repositories)
    # Print the identical string representation of the policy.
    print(general_policy)
    # Visit the policy
    MyGeneralPolicyVisitor().visit(general_policy)
    # Test whether the policy is structurally terminating
    assert(general_policy.is_terminating(repositories))
    # Test whether the policy solves the generalized state space.
    assert(general_policy.solves(knowledge_base.get_generalized_state_space(), denotation_repositories) == general_policies.SolvabilityStatus.SOLVED)

    # Create the policy from above off-the-shelf.
    general_policy = general_policies.GeneralPolicyFactory.get_or_create_general_policy_gripper(knowledge_base.get_domain(), repositories, dl_repositories)
    # Print the identical string representation of the policy.
    print(general_policy)
    # Visit the policy
    MyGeneralPolicyVisitor().visit(general_policy)
    # Test whether the policy is structurally terminating
    assert(general_policy.is_terminating(repositories))
    # Test whether the policy solves the generalized state space.
    assert(general_policy.solves(knowledge_base.get_generalized_state_space(), denotation_repositories) == general_policies.SolvabilityStatus.SOLVED)

if __name__ == "__main__":
    main()



""" This example illustrates the interface for the iterative width search algorithm and utilities.
"""

import pymimir.advanced.formalism as formalism
import pymimir.advanced.search as search

from pathlib import Path

ROOT_DIR = (Path(__file__).parent.parent.parent).absolute()


class CustomBrFSEventHandler(search.IBrFSEventHandler):
    """A custom event handler to react on events during iw search.
    
    It is useful to collect custom statistics or create a search tree.
    """
    def __init__(self, problem: formalism.Problem):
        super().__init__()

        self.problem = problem
        self.statistics = search.BrFSStatistics()

    def on_expand_state(self, state : search.State):
        pass

    def on_expand_goal_state(self, state : search.State):
        pass

    def on_generate_state(self, state : search.State, action : formalism.GroundAction, action_cost: float, successor_state: search.State):
        pass

    def on_generate_state_in_search_tree(self, state : search.State, action : formalism.GroundAction, action_cost: float, successor_state: search.State):
        print("Generated state:", successor_state.to_string(self.problem))

    def on_generate_state_not_in_search_tree(self, state : search.State, action : formalism.GroundAction, action_cost: float, successor_state: search.State):
        pass

    def on_finish_g_layer(self, g_value : int):
        pass

    def on_start_search(self, start_state : search.State):
        pass

    def on_end_search(self, num_reached_fluent_atoms : int, num_reached_derived_atoms: int, num_bytes_for_problem: int, num_bytes_for_nodes: int, num_states: int, num_nodes: int, num_actions: int, num_axioms: int):
        pass

    def on_solved(self, plan: search.Plan):
        pass

    def on_unsolvable(self):
        pass

    def on_exhausted(self):
        pass

    def get_statistics(self):
        return self.statistics


def main():
    domain_filepath = str(ROOT_DIR / "data" / "gripper" / "domain.pddl")
    problem_filepath = str(ROOT_DIR / "data" / "gripper" / "p-2-0.pddl")

    search_context = search.SearchContext.create(domain_filepath, problem_filepath)

    initial_state, _ = search_context.get_state_repository().get_or_create_initial_state()

    iw_event_handler = search.DefaultIWEventHandler(search_context.get_problem())
    brfs_event_handler = CustomBrFSEventHandler(search_context.get_problem())
    goal_strategy = search.ProblemGoalStrategy.create(search_context.get_problem())

    result = search.find_solution_iw(search_context, initial_state, 3, iw_event_handler, brfs_event_handler, goal_strategy)

    assert(result.status == search.SearchStatus.SOLVED)
    assert(len(result.plan) == 5)


if __name__ == "__main__":
    main()



""" This example illustrates the interface of a generic dynamic graph taking arbitrary python tuples as vertex and edge property.
    Dynamic graph provide the additional functionality of vertex and edge removal at the cost of less efficient underlying data structures.
"""

import pymimir.advanced.graphs as graphs


def main():
    graph = graphs.DynamicPyGraph()
    vx = graph.add_vertex(("to_be_removed",))
    v1 = graph.add_vertex((0,0))
    v2 = graph.add_vertex((0,1))
    v3 = graph.add_vertex((1,1))
    v4 = graph.add_vertex((1,0))
    graph.add_undirected_edge(v1, v2, ("a",1))
    graph.add_undirected_edge(v2, v3, ("b",2))
    graph.add_undirected_edge(v3, v4, ("c",3))
    graph.add_undirected_edge(v4, v1, ("d",4))
    graph.remove_vertex(vx)
    assert(v1 == 1)
    assert(v2 == 2)
    assert(v3 == 3)
    assert(v4 == 4)

    # Print dot representation of the graph.

    print(graph)

    # Check properties of the graph.

    assert(graphs.is_undirected(graph))
    assert(graphs.is_loopless(graph))
    assert(not graphs.is_acyclic(graph))
    assert(not graphs.is_multi(graph))

    # Run basic graph algorithms with switchable edge direction

    edge_weights = {e_idx : 1 for e_idx in graph.get_edge_indices()}
    source_indices = [1]

    graph.compute_strong_components()

    graph.compute_forward_breadth_first_search(source_indices)
    graph.compute_forward_depth_first_search(source_indices)
    graph.compute_forward_floyd_warshall_all_pairs_shortest_paths(edge_weights)
    graph.compute_forward_dijkstra_shortest_paths(edge_weights, source_indices)
    try:
        graph.compute_forward_topological_sort()  # graph is cyclic => this will throw ValueError
    except ValueError:
        pass
    
    graph.compute_backward_breadth_first_search(source_indices)
    graph.compute_backward_depth_first_search(source_indices)
    graph.compute_backward_floyd_warshall_all_pairs_shortest_paths(edge_weights)
    graph.compute_backward_dijkstra_shortest_paths(edge_weights, source_indices)
    try:
        graph.compute_backward_topological_sort()  # graph is cyclic => this will throw ValueError
    except ValueError:
        pass

    # Iterate over vertices
    for v_idx, vertex in graph.get_vertices().items():
        # Access vertex properties through generic member
        vertex.get_index()      
        vertex.get_property_0()  # python tuple

        # Iterate over adjacent structures
        for adj_vertex in graph.get_forward_adjacent_vertices(vertex.get_index()):
            pass

        for adj_vertex in graph.get_backward_adjacent_vertices(vertex.get_index()):
            pass 

        for adj_edge in graph.get_forward_adjacent_edges(vertex.get_index()):
            pass 

        for adj_edge in graph.get_backward_adjacent_edges(vertex.get_index()):
            pass 

        for adj_v_idx in graph.get_forward_adjacent_vertex_indices(vertex.get_index()):
            pass 

        for adj_v_idx in graph.get_backward_adjacent_vertex_indices(vertex.get_index()):
            pass 

        for adj_e_idx in graph.get_forward_adjacent_edge_indices(vertex.get_index()):
            pass 

        for adj_e_idx in graph.get_backward_adjacent_edge_indices(vertex.get_index()):
            pass
            
    # Iterate over edges
    for e_idx, edge in graph.get_edges().items():
        # Access edge properties through generic member
        edge.get_index()  
        edge.get_source()   
        edge.get_target()   
        edge.get_property_0()  # python tuple



if __name__ == "__main__":
    main()